from functools import partial

from PySide6.QtCore import Qt, Slot, QObject, QDateTime
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu, QDialog
from jal.db.helpers import load_icon
from jal.ui.reports.ui_holdings_report import Ui_HoldingsWidget
from jal.reports.reports import Reports
from jal.db.asset import JalAsset
from jal.db.holdings_model import HoldingsModel
from jal.widgets.mdi import MdiWidget
from jal.db.tax_estimator import TaxEstimator
from jal.widgets.price_chart import ChartWindow
from jal.widgets.selection_dialog import SelectTagDialog

JAL_REPORT_CLASS = "HoldingsReport"


# ----------------------------------------------------------------------------------------------------------------------
class HoldingsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Holdings")
        self.window_class = "HoldingsReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class HoldingsReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_HoldingsWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Holdings")

        # Add available groupings
        self.ui.GroupCombo.addItem(self.tr("Currency - Account"), "currency_id;account_id")
        self.ui.GroupCombo.addItem(self.tr("Asset"), "asset_id")
        self.ui.GroupCombo.addItem(self.tr("Country"), "country_id")
        self.ui.GroupCombo.addItem(self.tr("Tag"), "tag")
        self.ui.GroupCombo.addItem("None", "")

        self.holdings_model = HoldingsModel(self.ui.HoldingsTreeView)
        self.ui.HoldingsTreeView.setModel(self.holdings_model)
        self.ui.HoldingsTreeView.setContextMenuPolicy(Qt.CustomContextMenu)

        # Setup holdings parameters
        current_time = QDateTime.currentDateTime()
        current_time.setTimeSpec(Qt.UTC)  # We use UTC everywhere so need to force TZ info
        self.ui.HoldingsDate.setDateTime(current_time)
        self.ui.HoldingsCurrencyCombo.setIndex(JalAsset.get_base_currency())

        self.connect_signals_and_slots()
        self.updateReport()

    def connect_signals_and_slots(self):
        self.ui.HoldingsDate.dateChanged.connect(self.updateReport)
        self.ui.HoldingsCurrencyCombo.changed.connect(self.updateReport)
        self.ui.HoldingsTreeView.customContextMenuRequested.connect(self.onHoldingsContextMenu)
        self.ui.GroupCombo.currentIndexChanged.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.HoldingsTreeView.model()))

    @Slot()
    def updateReport(self):
        self.ui.HoldingsTreeView.model().updateView(currency_id = self.ui.HoldingsCurrencyCombo.selected_id,
                                                    date = self.ui.HoldingsDate.date(),
                                                    grouping = self.ui.GroupCombo.currentData())

    @Slot()
    def onHoldingsContextMenu(self, pos):
        index = self.ui.HoldingsTreeView.indexAt(pos)
        contextMenu = QMenu(self.ui.HoldingsTreeView)
        actionShowChart = QAction(icon=load_icon("chart.png"), text=self.tr("Show Price Chart"), parent=self.ui.HoldingsTreeView)
        actionShowChart.triggered.connect(partial(self.showPriceChart, index))
        contextMenu.addAction(actionShowChart)
        tax_submenu = contextMenu.addMenu(load_icon("tax.png"), self.tr("Estimate tax"))
        actionEstimateTaxPt = QAction(icon=load_icon("pt.png"), text=self.tr("Portugal"), parent=self.ui.HoldingsTreeView)
        actionEstimateTaxPt.triggered.connect(partial(self.estimateRussianTax, index, 'pt'))
        tax_submenu.addAction(actionEstimateTaxPt)
        actionEstimateTaxRu = QAction(icon=load_icon("ru.png"), text=self.tr("Russia"), parent=self.ui.HoldingsTreeView)
        actionEstimateTaxRu.triggered.connect(partial(self.estimateRussianTax, index, 'ru'))
        tax_submenu.addAction(actionEstimateTaxRu)
        contextMenu.addSeparator()
        actionSetTag = QAction(icon=load_icon("tag.png"), text=self.tr("Set asset tag"), parent=self.ui.HoldingsTreeView)
        actionSetTag.triggered.connect(partial(self.setTag, index))
        contextMenu.addAction(actionSetTag)
        contextMenu.addSeparator()
        actionExpandAll = QAction(text=self.tr("Expand all"),parent=self.ui.HoldingsTreeView)
        actionExpandAll.triggered.connect(self.ui.HoldingsTreeView.expandAll)
        contextMenu.addAction(actionExpandAll)
        actionCollapseAll = QAction(text=self.tr("Collapse all"), parent=self.ui.HoldingsTreeView)
        actionCollapseAll.triggered.connect(self.ui.HoldingsTreeView.collapseAll)
        contextMenu.addAction(actionCollapseAll)
        contextMenu.popup(self.ui.HoldingsTreeView.viewport().mapToGlobal(pos))

    @Slot()
    def showPriceChart(self, index):
        model = index.model()
        account, asset, currency, asset_qty = model.get_data_for_tax(index)
        self._parent.mdi_area().addSubWindow(ChartWindow(account, asset, currency, asset_qty))

    @Slot()
    def estimateRussianTax(self, index, country_code):
        model = index.model()
        account, asset, currency, asset_qty = model.get_data_for_tax(index)
        self._parent.mdi_area().addSubWindow(TaxEstimator(country_code, account, asset, asset_qty), size=(1000, 300))

    @Slot()
    def setTag(self, index):
        model = index.model()
        account, asset_id, currency, asset_qty = model.get_data_for_tax(index)
        asset = JalAsset(asset_id)
        dialog = SelectTagDialog(
            parent=self, description=self.tr("Select tag for {} ({}): ").format(asset.symbol(currency),asset.name()))
        if dialog.exec() != QDialog.Accepted:
            return
        asset.set_tag(dialog.selected_id)



import logging
from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex, QMimeData, QByteArray, QDataStream, QIODevice
from PySide6.QtSql import QSqlTableModel, QSqlRelationalTableModel
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QHeaderView, QMessageBox, QAbstractItemView
from jal.db.db import JalDB, JalSqlError


# ----------------------------------------------------------------------------------------------------------------------
class AbstractReferenceListModel(QSqlRelationalTableModel, JalDB):
    @property
    def completion_model(self):
        return self._completion_model

    def __init__(self, table, parent_view, **kwargs):
        self._view = parent_view
        self._table = table
        self._columns = []
        self._deleted_rows = []
        self._default_name = "name"
        self._group_by = None
        self._filter_by = ''
        self._filter_value = None
        self._sort_by = None
        self._hidden = []
        self._stretch = None
        self._default_values = {}   # To fill in default values for fields allowed to be NULL
        super().__init__(parent=parent_view, db=self.connection())
        self.setJoinMode(QSqlRelationalTableModel.LeftJoin)
        self.setTable(self._table)
        self.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.select()
        # This is auxiliary 'plain' model of the same table - to be given as QCompleter source of data
        self._completion_model = QSqlTableModel(parent=parent_view, db=self.connection())
        self._completion_model.setTable(self._table)
        self._completion_model.select()

    @property
    def group_by(self):
        return self._group_by

    def fieldIndex(self, field):
        column_data = [i for i, column in enumerate(self._columns) if column[0] == field]
        if len(column_data) > 0:
            return column_data[0]
        else:
            return -1

    def headerData(self, section, orientation=Qt.Horizontal, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self._columns[section][1]
        return None

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.FontRole and (index.row() in self._deleted_rows):  # Strike-out deleted items
            font = QFont()
            font.setStrikeOut(True)
            return font
        return super().data(index, role)

    def configureView(self):
        self.setSorting()
        self.hideColumns()
        self.setStretching()
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)

    def setSorting(self):
        if self._sort_by:
            self.setSort(self.fieldIndex(self._sort_by), Qt.AscendingOrder)

    def hideColumns(self):
        for column_name in self._hidden:
            self._view.setColumnHidden(self.fieldIndex(column_name), True)

    def setStretching(self):
        if self._stretch:
            self._view.horizontalHeader().setSectionResizeMode(self.fieldIndex(self._stretch), QHeaderView.Stretch)

    def getId(self, index):
        return self.record(index.row()).value('id')

    def getName(self, index):
        return self.getFieldValue(self.getId(index), self._default_name)

    def getFieldValue(self, item_id, field_name):
        return self._read(f"SELECT {field_name} FROM {self._table} WHERE id=:id", [(":id", item_id)])

    def addElement(self, index, in_group=0):
        row = index.row() if index.isValid() else 0
        assert self.insertRows(row, 1)
        new_record = self.record()
        if in_group != 0:
            new_record.setValue(self.fieldIndex(self._group_by), in_group)   # by index as it is lookup field
        for field in self._default_values:
            new_record.setValue(self.fieldIndex(field), self._default_values[field])
        if self._filter_by:
            new_record.setValue(self.fieldIndex(self._filter_by), self._filter_value)
        self.setRecord(row, new_record)

    def removeElement(self, index):
        if index.isValid():
            row = index.row()
        else:
            return
        assert self.removeRow(row)
        self._deleted_rows.append(row)

    def submitAll(self):
        result = super().submitAll()
        if result:
            self._deleted_rows = []
        else:
            error_code = self.lastError().nativeErrorCode()
            null_pfx = "NOT NULL constraint failed: " + self.tableName() + "."
            if error_code == '1299' and self.lastError().databaseText().startswith(null_pfx):
                    field_name = self.lastError().databaseText()[len(null_pfx):]
                    header_title = self.tableName() + ":" + self.headerData(self.fieldIndex(field_name))
                    QMessageBox().warning(self._view, self.tr("Data are incomplete"),
                                          self.tr("Column has no valid value: " + header_title), QMessageBox.Ok)
            else:
                error = JalSqlError(self.lastError().databaseText())
                if error.custom():
                    error.show()
                else:
                    logging.fatal(self.tr("Submit failed: ") + error.message())
        return result

    def revertAll(self):
        self._deleted_rows = []
        super().revertAll()

    def locateItem(self, item_id, use_filter=''):
        if use_filter:
            use_filter = f"WHERE {use_filter}"
        row = self._read(f"SELECT row_number FROM ("
                         f"SELECT ROW_NUMBER() OVER (ORDER BY {self._default_name}) AS row_number, id "
                         f"FROM {self._table} {use_filter}) WHERE id=:id", [(":id", item_id)])
        if row is None:
            return QModelIndex()
        return self.index(row - 1, 0)

    def updateItemType(self, index, new_type):
        id = self.getId(index)
        self._exec(f"UPDATE {self._table} SET {self._group_by}=:new_type WHERE id=:id",
                   [(":new_type", new_type), (":id", id)])

    def filterBy(self, field_name, value):
        self._filter_by = field_name
        self._filter_value = value
        self.setFilter(f"{self._table}.{field_name} = {value}")

    # returns group id for given item
    def getGroupId(self, item_id: int) -> int:
        group_id = self._read(f"SELECT {self._group_by} FROM {self._table} WHERE id=:id", [(":id", item_id)])
        group_id = 0 if group_id is None else group_id
        return group_id


# ----------------------------------------------------------------------------------------------------------------------
class SqlTreeModel(QAbstractItemModel, JalDB):
    ROOT_PID = 0
    DRAG_DROP_MIME_TYPE = "application/vnd.tree_item"

    @property
    def completion_model(self):
        return self._completion_model

    def __init__(self, table, parent_view):
        super().__init__(parent=parent_view)
        self._table = table
        self._columns = []
        self._drag_and_drop = False  # This is required to prevent deletion of initial element after drag&drop movement
        self._view = parent_view
        self._default_name = "name"
        self._stretch = None
        self._sort_by = None
        self._filter_text = ''
        # This is auxiliary 'plain' model of the same table - to be given as QCompleter source of data
        self._completion_model = QSqlTableModel(parent=parent_view, db=self.connection())
        self._completion_model.setTable(self._table)
        self._completion_model.select()

    def index(self, row, column, parent=None):
        if parent is None:
            return QModelIndex()
        if not parent.isValid():
            parent_id = self.ROOT_PID
        else:
            parent_id = parent.internalId()
        order_by = f"ORDER BY {self._sort_by}" if self._sort_by is not None else ''
        if not self._filter_text:
            child_id = self._read(f"SELECT id FROM {self._table} WHERE pid=:pid {order_by} LIMIT 1 OFFSET :row_n",
                                  [(":pid", parent_id), (":row_n", row)])
        else:  # display a plain list in a filter mode
            child_id = self._read(f"SELECT id FROM {self._table} WHERE {self._filter_text} {order_by} "
                                  "LIMIT 1 OFFSET :row_n", [(":row_n", row)])
        if child_id:
            return self.createIndex(row, column, id=child_id)
        return QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()
        child_id = index.internalId()
        parent_id = self._read(f"SELECT pid FROM {self._table} WHERE id=:id", [(":id", child_id)])
        if parent_id == self.ROOT_PID:
            return QModelIndex()
        order_by = f"ORDER BY {self._sort_by}" if self._sort_by is not None else ''
        row = self._read(f"SELECT row_number FROM ("
                         f"SELECT ROW_NUMBER() OVER ({order_by}) AS row_number, id, pid "
                         f"FROM {self._table} WHERE pid IN (SELECT pid FROM {self._table} WHERE id=:id)) "
                         f"WHERE id=:id", [(":id", parent_id)])
        return self.createIndex(row-1, 0, id=parent_id)

    def rowCount(self, parent=None):
        if not parent.isValid():
            parent_id = self.ROOT_PID
        else:
            parent_id = parent.internalId()
        if not self._filter_text:
            count = self._read(f"SELECT COUNT(id) FROM {self._table} WHERE pid=:pid", [(":pid", parent_id)])
        else:
            if parent_id:  # no children if we are in filter mode and display a plain list
                return 0
            count = self._read(f"SELECT COUNT(id) FROM {self._table} WHERE {self._filter_text}")
        if count:
            return int(count)
        else:
            return 0

    def columnCount(self, parent=None):
        return len(self._columns)

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.ItemIsDropEnabled
        return Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled | Qt.ItemFlag.ItemIsDragEnabled | super().flags(index)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        item_id = index.internalId()
        if role == Qt.DisplayRole:
            col = index.column()
            if (col >= 0) and (col < len(self._columns)):
                return self._read(f"SELECT {self._columns[col][0]} FROM {self._table} WHERE id=:id",
                                  [(":id", item_id)])
            else:
                return None
        return None

    def setData(self, index, value, role=Qt.EditRole):
        if role != Qt.EditRole:
            return False
        if not index.isValid():
            return False
        item_id = index.internalId()
        col = index.column()
        self.connection().transaction()
        _ = self._exec(f"UPDATE {self._table} SET {self._columns[col][0]}=:value WHERE id=:id",
                       [(":id", item_id), (":value", value)])
        self.dataChanged.emit(index, index, Qt.DisplayRole | Qt.EditRole)
        return True

    def fieldIndex(self, field):
        column_data = [i for i, column in enumerate(self._columns) if column[0] == field]
        if len(column_data) > 0:
            return column_data[0]
        else:
            return -1

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self._columns[section][1]
        return None

    def supportedDragActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction

    def supportedDropActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction

    def mimeTypes(self) -> list:
        return [self.DRAG_DROP_MIME_TYPE]

    # This method encodes tree element 'id' field and returns it as QMimeData for Drag&Drop operation
    def mimeData(self, indexes: list) -> QMimeData:
        item_data = QMimeData()
        encoded_data = QByteArray()
        stream = QDataStream(encoded_data, QIODevice.WriteOnly)
        for index in indexes:
            stream.writeUInt64(index.internalId())
        item_data.setData(self.DRAG_DROP_MIME_TYPE, encoded_data)
        return item_data

    def canDropMimeData(self, data: QMimeData, action: Qt.DropAction, row: int, column: int, parent: QModelIndex) -> bool:
        if action != Qt.DropAction.MoveAction or not data.hasFormat(self.DRAG_DROP_MIME_TYPE):
            return False
        return True

    def dropMimeData(self, data: QMimeData, action: Qt.DropAction, row: int, column: int, parent: QModelIndex) -> bool:
        if not self.canDropMimeData(data, action, row, column, parent):
            return False
        encoded_data = data.data(self.DRAG_DROP_MIME_TYPE)
        stream = QDataStream(encoded_data, QIODevice.ReadOnly)
        item_id = stream.readUInt64()
        self.connection().transaction()
        if parent.isValid():
            self._exec(f"UPDATE {self._table} SET pid=:pid WHERE id=:id",
                       [(":id", item_id), (":pid", parent.internalId())])
        else:
            self._exec(f"UPDATE {self._table} SET pid=0 WHERE id=:id", [(":id", item_id)])
        self._drag_and_drop = True
        return True

    def configureView(self):
        self.setStretching()
        font = self._view.header().font()
        font.setBold(True)
        self._view.header().setFont(font)
        self._view.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self._view.setDragEnabled(True)
        self._view.setAcceptDrops(True)
        self._view.setDropIndicatorShown(True)
        self._view.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self._view.setDefaultDropAction(Qt.DropAction.MoveAction)

    def setStretching(self):
        if self._stretch:
            self._view.header().setSectionResizeMode(self.fieldIndex(self._stretch), QHeaderView.Stretch)

    def getId(self, index):
        if not index.isValid():
            return None
        return index.internalId()

    def getName(self, index):
        return self.getFieldValue(self.getId(index), self._default_name)

    def getFieldValue(self, item_id, field_name):
        return self._read(f"SELECT {field_name} FROM {self._table} WHERE id=:id", [(":id", item_id)])

    def deleteWithChilderen(self, parent_id: int) -> None:
        query = self._exec(f"SELECT id FROM {self._table} WHERE pid=:pid", [(":pid", parent_id)])
        while query.next():
            self.deleteWithChilderen(query.value(0))
        _ = self._exec(f"DELETE FROM {self._table} WHERE id=:id", [(":id", parent_id)])

    def insertRows(self, row, count, parent=None):
        if parent is None:
            return False
        if not parent.isValid():
            parent_id = self.ROOT_PID
        else:
            parent_id = parent.internalId()

        self.beginInsertRows(parent, row, row + count - 1)
        self.connection().transaction()
        _ = self._exec(f"INSERT INTO {self._table}(pid, {self._default_name}) VALUES (:pid, '')", [(":pid", parent_id)])
        self.endInsertRows()
        self.layoutChanged.emit()
        return True

    def removeRows(self, row, count, parent=None):
        if self._drag_and_drop:  # This is an automatically triggered action - keep the element but refresh the view
            self._drag_and_drop = False
            self.dataChanged.emit(QModelIndex(), QModelIndex(), Qt.DisplayRole)  # the call is required to enable commit/rollback buttons in UI
            self.layoutChanged.emit()
            return True
        if parent is None:
            return False
        if not parent.isValid():
            parent_id = self.ROOT_PID
        else:
            parent_id = parent.internalId()

        self.beginRemoveRows(parent, row, row + count - 1)
        self.connection().transaction()
        query = self._exec(f"SELECT id FROM {self._table} WHERE pid=:pid LIMIT :row_c OFFSET :row_n",
                           [(":pid", parent_id), (":row_c", count), (":row_n", row)])
        while query.next():
            self.deleteWithChilderen(query.value(0))
        self.endRemoveRows()
        self.layoutChanged.emit()
        return True

    def addElement(self, index, in_group=0):  # in_group is used for plain model only, not tree
        row = index.row()
        assert self.insertRows(row, 1, index.parent())

    def addChildElement(self, index):
        assert self.insertRows(0, 1, index)
        self._view.expand(index)

    def removeElement(self, index):
        row = index.row()
        assert self.removeRows(row, 1, index.parent())

    def submitAll(self):
        _ = self._exec("COMMIT")
        self.layoutChanged.emit()
        return True

    def revertAll(self):
        _ = self._exec("ROLLBACK")
        self.layoutChanged.emit()

    # expand all parent elements for tree element with given index
    def expand_parent(self, index):
        parent = index.parent()
        if parent.isValid():
            self.expand_parent(parent)
        self._view.expand(index)

    # find item by ID and make it selected in associated self._view
    def locateItem(self, item_id):
        order_by = f"ORDER BY {self._sort_by}" if self._sort_by is not None else ''
        row = self._read(f"SELECT row_number FROM ("
                         f"SELECT ROW_NUMBER() OVER ({order_by}) AS row_number, id, pid "
                         f"FROM {self._table} WHERE pid IN (SELECT pid FROM {self._table} WHERE id=:id)) "
                         f"WHERE id=:id", [(":id", item_id)])
        if row is None:
            return
        item_idx = self.createIndex(row-1, 0, id=item_id)
        self.expand_parent(item_idx)
        self._view.setCurrentIndex(item_idx)

    def setFilter(self, text):
        self._filter_text = text
        self.layoutChanged.emit()



from functools import partial

from PySide6.QtCore import Slot, QObject
from jal.ui.reports.ui_deals_report import Ui_DealsReportWidget
from jal.reports.reports import Reports
from jal.widgets.mdi import MdiWidget
from jal.db.trades_model import ClosedTradesModel

JAL_REPORT_CLASS = "DealsReport"


# ----------------------------------------------------------------------------------------------------------------------
class DealsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.name = self.tr("Deals by Account")
        self.window_class = "DealsReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class DealsReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_DealsReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Deals")

        # Add available groupings
        self.ui.GroupCombo.addItem(self.tr("<None>"), "")
        self.ui.GroupCombo.addItem(self.tr("Asset"), "symbol")
        self.ui.GroupCombo.addItem(self.tr("Close"), "close_date")
        self.ui.GroupCombo.addItem(self.tr("Asset - Open - Close"), "symbol;open_date;close_date")
        self.ui.GroupCombo.addItem(self.tr("Open - Close"), "open_date;close_date")
        self.ui.GroupCombo.addItem(self.tr("Close - Open"), "close_date;open_date")

        self.trades_model = ClosedTradesModel(self.ui.ReportTreeView)
        self.ui.ReportTreeView.setModel(self.trades_model)

        self.connect_signals_and_slots()

    def connect_signals_and_slots(self):
        self.ui.ReportAccountBtn.changed.connect(self.updateReport)
        self.ui.ReportRange.changed.connect(self.updateReport)
        self.ui.GroupCombo.currentIndexChanged.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTreeView.model()))

    @Slot()
    def updateReport(self):
        self.ui.ReportTreeView.model().updateView(account_id=self.ui.ReportAccountBtn.account_id,
                                                  dates=self.ui.ReportRange.getRange(),
                                                  grouping=self.ui.GroupCombo.currentData())



from decimal import Decimal
from datetime import datetime

from jal.db.operations import Dividend
from jal.data_export.taxes import TaxReport


class TaxesPortugal(TaxReport):
    currency_name = 'EUR'
    country_name = 'portugal'

    def __init__(self):
        super().__init__()
        self._processed_trade_qty = {}  # It will handle {trade_id: qty} records to keep track of already processed qty
        self.reports = {
            "Dividends": (self.prepare_dividends, "tax_prt_dividends.json"),
            "Shares": (self.prepare_stocks_and_etf, "tax_prt_shares.json")
        }

    def prepare_dividends(self):
        dividends_report = []
        dividends = self.dividends_list()
        dividends = [x for x in dividends if x.amount(self.account_currency.id()) > 0]  # Skip negative dividends
        for dividend in dividends:
            country = dividend.asset().country()
            note = ''
            if dividend.subtype() == Dividend.StockDividend:
                note = "Stock dividend"
            if dividend.subtype() == Dividend.StockVesting:
                note = "Stock vesting"
            line = {
                'report_template': "dividend",
                'payment_date': dividend.timestamp(),
                'symbol': dividend.asset().symbol(self.account_currency.id()),
                'full_name': dividend.asset().name(),
                'isin': dividend.asset().isin(),
                'amount': dividend.amount(self.account_currency.id()),
                'tax': dividend.tax(),
                'rate': self.account_currency.quote(dividend.timestamp(), self._currency_id)[1],
                'country': country.name(language='en'),
                'tax_treaty': "Y" if self.has_tax_treaty_with(country.code()) else "N",
                'amount_eur': round(dividend.amount(self._currency_id), 2),
                'tax_eur': round(dividend.tax(self._currency_id), 2),
                'note': note
            }
            dividends_report.append(line)
        self.insert_totals(dividends_report, ["amount", "amount_eur", "tax", "tax_eur"])
        return dividends_report

    def inflation(self, timestamp: int) -> Decimal:
        year = datetime.utcfromtimestamp(timestamp).strftime('%Y')
        inflation_coefficients = self._parameters['currency_devaluation']
        try:
            coefficient = Decimal(inflation_coefficients[year])
        except KeyError:
            coefficient = Decimal('1')
        return coefficient

# ----------------------------------------------------------------------------------------------------------------------
    def prepare_stocks_and_etf(self):
        deals_report = []
        ns = not self.use_settlement
        trades = self.shares_trades_list()
        for trade in trades:
            if ns:
                os_rate = self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1]
            else:
                os_rate = self.account_currency.quote(trade.open_operation().settlement(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().settlement(), self._currency_id)[1]
            if trade.qty() >= Decimal('0'):  # Long trade
                note = ''
                income = round(trade.close_amount(no_settlement=ns), 2)
                income_eur = round(trade.close_amount(self._currency_id, no_settlement=ns), 2)
                spending = round(trade.open_amount(no_settlement=ns), 2) + trade.fee()
                spending_eur = round(trade.open_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2)
            else:  # Short trade
                # Check were there any dividends during short position holding
                short_dividend_eur = Decimal('0')
                dividends = Dividend.get_list(self.account.id(), subtype=Dividend.Dividend)
                dividends = [x for x in dividends if
                             trade.open_operation().settlement() <= x.ex_date() <= trade.close_operation().settlement()]
                for dividend in dividends:
                    short_dividend_eur += dividend.amount(self._currency_id)
                note = f"Dividend withheld: {short_dividend_eur} EUR" if short_dividend_eur > Decimal('0') else ''
                income = round(trade.open_amount(no_settlement=ns), 2)
                income_eur = round(trade.open_amount(self._currency_id, no_settlement=ns), 2)
                spending = round(trade.close_amount(no_settlement=ns), 2) + trade.fee() + short_dividend_eur
                spending_eur = round(trade.close_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2)
            inflation = self.inflation(trade.open_operation().timestamp())
            if inflation != Decimal('1'):
                spending_eur *= inflation
                note = f"Inflation coefficient: {inflation:.2f}\n" + note
            line = {
                'report_template': "trade",
                'symbol': trade.asset().symbol(self.account_currency.id()),
                'isin': trade.asset().isin(),
                'qty': trade.qty(),
                'o_type': "Buy" if trade.qty() >= Decimal('0') else "Sell",
                'o_number': trade.open_operation().number(),
                'o_date': trade.open_operation().timestamp(),
                'o_rate': self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1],
                'os_date': trade.open_operation().settlement(),
                'os_rate': os_rate,
                'o_price': trade.open_operation().price(),
                'o_amount': round(trade.open_amount(no_settlement=ns), 2),
                'o_amount_eur': round(trade.open_amount(self._currency_id, no_settlement=ns), 2),
                'o_fee': trade.open_fee(),
                'o_fee_eur': round(trade.open_fee(self._currency_id), 2),
                'c_type': "Sell" if trade.qty() >= Decimal('0') else "Buy",
                'c_number': trade.close_operation().number(),
                'c_date': trade.close_operation().timestamp(),
                'c_rate': self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1],
                'cs_date': trade.close_operation().settlement(),
                'cs_rate': cs_rate,
                'c_price': trade.close_operation().price(),
                'c_amount': round(trade.close_amount(no_settlement=ns), 2),
                'c_amount_eur': round(trade.close_amount(self._currency_id, no_settlement=ns), 2),
                'c_fee': trade.close_fee(),
                'c_fee_eur': round(trade.close_fee(self._currency_id), 2),
                'income_eur': income_eur,
                'spending_eur': spending_eur,
                'profit': income - spending,
                'profit_eur': income_eur - spending_eur,
                'note': note
            }
            deals_report.append(line)
        self.insert_totals(deals_report, ["income_eur", "spending_eur", "profit_eur", "profit"])
        return deals_report



from jal.db.db import JalDB
from jal.constants import AssetData

class JalTag(JalDB):
    db_cache = []

    def __init__(self, tag_id: int = 0) -> None:
        super().__init__(cached=True)
        if not JalTag.db_cache:
            self._fetch_data()
        self._id = tag_id
        try:
            self._data = [x for x in self.db_cache if x['id'] == self._id][0]
        except IndexError:
            self._data = None
        self._name = self._data['tag'] if self._data is not None else ''

    def invalidate_cache(self):
        self._fetch_data()

    # JalCountry maintains single cache available for all instances
    @classmethod
    def class_cache(cls) -> True:
        return True

    def _fetch_data(self):
        JalTag.db_cache = []
        query = self._exec("SELECT * FROM tags ORDER BY id")
        while query.next():
            JalTag.db_cache.append(self._read_record(query, named=True))

    def id(self) -> int:
        return self._id

    # Returns country name in given language or in current interface language if no argument is given
    def name(self) -> str:
        return self._name

    def replace_with(self, new_id):
        self._exec("UPDATE action_details SET tag_id=:new_id WHERE tag_id=:old_id",
                   [(":new_id", new_id), (":old_id", self._id)])
        self._exec("UPDATE asset_data SET value=:new_id WHERE datatype=:tag AND value=:old_id",
                   [(":tag", AssetData.Tag), (":new_id", str(new_id)), (":old_id", self._id)])
        self._exec("DELETE FROM tags WHERE id=:old_id", [(":old_id", self._id)], commit=True)
        JalDB().invalidate_cache()  # Full DB as it impacts JalAsset cache also
        self._id = 0



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QHeaderView
from PySide6.QtGui import QFont
from jal.db.db import JalModel


class JalViewModel(JalModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = []
        self.deleted = []
        self._view = parent_view

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def removeRow(self, row, parent=None):
        self.deleted.append(row)
        super().removeRow(row)

    def submitAll(self):
        result = super().submitAll()
        if result:
            self.deleted = []
        return result

    def revertAll(self):
        self.deleted = []
        super().revertAll()

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.FontRole and (index.row() in self.deleted):
            font = QFont()
            font.setStrikeOut(True)
            return font
        return super().data(index, role)

    def configureView(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 100)
        self._view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)



from PySide6.QtCore import Signal, Slot
from PySide6.QtWidgets import QWidget, QStackedWidget, QMessageBox

from jal.widgets.corporate_action_widget import CorporateActionWidget
from jal.widgets.dividend_widget import DividendWidget
from jal.widgets.income_spending_widget import IncomeSpendingWidget
from jal.widgets.trade_widget import TradeWidget
from jal.widgets.transfer_widget import TransferWidget
from jal.db.operations import LedgerTransaction


class JalOperationsTabs(QStackedWidget):
    dbUpdated = Signal()

    def __init__(self, parent):
        super().__init__(parent)
        self.widgets = {LedgerTransaction.NA: QWidget(self),
                        LedgerTransaction.IncomeSpending: IncomeSpendingWidget(self),
                        LedgerTransaction.Dividend: DividendWidget(self), LedgerTransaction.Trade: TradeWidget(self),
                        LedgerTransaction.Transfer: TransferWidget(self),
                        LedgerTransaction.CorporateAction: CorporateActionWidget(self)}
        for key, widget in self.widgets.items():
            if key != LedgerTransaction.NA:
                widget.dbUpdated.connect(self.dbUpdated)
            self.addWidget(widget)
        self.setCurrentIndex(0)

    # Returns a dictionary of {op_type, op_name} of operations that widget is able to handle
    def get_operations_list(self) -> dict:
        operations = {}
        for key, widget in self.widgets.items():
            if hasattr(widget, "name"):
                operations[key] = widget.name
        return operations

    def _check_for_changes(self):
        for key, widget in self.widgets.items():
            if key == LedgerTransaction.NA:
                continue
            if widget.modified:
                reply = QMessageBox().warning(None, self.tr("You have unsaved changes"),
                                              widget.name +
                                              self.tr(" has uncommitted changes,\ndo you want to save it?"),
                                              QMessageBox.Yes, QMessageBox.No)
                if reply == QMessageBox.Yes:
                    widget.saveChanges()
                else:
                    widget.revertChanges()

    def show_operation(self, op_type, operation_id):
        self._check_for_changes()
        self.setCurrentIndex(op_type)
        if op_type != LedgerTransaction.NA:
            self.widgets[op_type].setId(operation_id)

    def new_operation(self, op_type, account_id):
        self._check_for_changes()
        self.widgets[op_type].createNew(account_id=account_id)
        self.setCurrentIndex(op_type)

    @Slot()
    def copy_operation(self):
        op_type = self.currentIndex()
        if op_type == LedgerTransaction.NA:
            return
        self._check_for_changes()
        self.widgets[op_type].copyNew()



from decimal import Decimal
from jal.constants import PredefinedAsset, PredefinedCategory
from jal.db.helpers import remove_exponent
from jal.db.operations import LedgerTransaction, Dividend, CorporateAction
from jal.db.asset import JalAsset
from jal.db.category import JalCategory
from jal.data_export.taxes import TaxReport


# -----------------------------------------------------------------------------------------------------------------------
class TaxesRussia(TaxReport):
    currency_name = 'RUB'
    country_name = "russia"

    CorpActionText = {
        CorporateAction.SymbolChange: "Смена символа {before} {old} -> {after} {new}",
        CorporateAction.Split: "Сплит {old} {before} в {after} {new}",
        CorporateAction.SpinOff: "Выделение компании {new} из {old}; доля выделяемого актива {share:.2f}%",
        CorporateAction.Merger: "Реорганизация компании, конвертация {share:.2f}% стоимости {before} {old} в {after} {new}",
        CorporateAction.Delisting: "Делистинг"
    }

    def __init__(self):
        super().__init__()
        self.broker_name = ''
        self.broker_iso_cc = "000"
        self._processed_trade_qty = {}  # It will handle {trade_id: qty} records to keep track of already processed qty
        self.reports = {
            "Дивиденды": (self.prepare_dividends, "tax_rus_dividends.json"),
            "Акции": (self.prepare_stocks_and_etf, "tax_rus_trades.json"),
            "Облигации": (self.prepare_bonds, "tax_rus_bonds.json"),
            "ПФИ": (self.prepare_derivatives, "tax_rus_derivatives.json"),
            "Криптовалюты": (self.prepare_crypto, "tax_rus_crypto.json"),
            "Корп.события": (self.prepare_corporate_actions, "tax_rus_corporate_actions.json"),
            "Комиссии": (self.prepare_broker_fees, "tax_rus_fees.json"),
            "Проценты": (self.prepare_broker_interest, "tax_rus_interests.json")
        }

    def prepare_dividends(self):
        dividends_report = []
        dividends = self.dividends_list()
        dividends = [x for x in dividends if x.amount(self.account_currency.id()) > 0]  # Skip negative dividends
        for dividend in dividends:
            country = dividend.asset().country()
            note = ''
            if dividend.subtype() == Dividend.StockDividend:
                note = "Дивиденд выплачен в натуральной форме (ценными бумагами)"
            if dividend.subtype() == Dividend.StockVesting:
                note = "Доход получен в натуральной форме (ценными бумагами)"
            tax_rub = dividend.tax(self._currency_id)
            tax2pay = Decimal('0.13') * dividend.amount(self._currency_id)
            if self.has_tax_treaty_with(country.code()):
                if tax2pay > tax_rub:
                    tax2pay = tax2pay - tax_rub
                else:
                    tax2pay = Decimal('0.0')
            line = {
                'report_template': "dividend",
                'payment_date': dividend.timestamp(),
                'symbol': dividend.asset().symbol(self.account_currency.id()),
                'full_name': dividend.asset().name(),
                'isin': dividend.asset().isin(),
                'amount': dividend.amount(self.account_currency.id()),
                'tax': dividend.tax(),
                'rate': self.account_currency.quote(dividend.timestamp(), self._currency_id)[1],
                'country': country.name(language='ru'),
                'country_iso': country.iso_code(),  # it is required for DLSG export
                'tax_treaty': "Да" if self.has_tax_treaty_with(country.code()) else "Нет",
                'amount_rub': round(dividend.amount(self._currency_id), 2),
                'tax_rub': round(dividend.tax(self._currency_id), 2),
                'tax2pay': round(tax2pay, 2),
                'note': note
            }
            dividends_report.append(line)
        self.insert_totals(dividends_report, ["amount", "amount_rub", "tax", "tax_rub", "tax2pay"])
        return dividends_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_trades_report(self, trades_list):
        deals_report = []
        ns = not self.use_settlement
        # Prepare list of dividends withdrawn from account (due to short trades)
        dividends_withdrawn = Dividend.get_list(self.account.id(), subtype=Dividend.Dividend)
        dividends_withdrawn = [x for x in dividends_withdrawn if self.year_begin <= x.timestamp() <= self.year_end]
        dividends_withdrawn = [x for x in dividends_withdrawn if x.amount() < Decimal('0')]
        for trade in trades_list:
            if ns:
                os_rate = self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1]
            else:
                os_rate = self.account_currency.quote(trade.open_operation().settlement(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().settlement(), self._currency_id)[1]
            if trade.qty() >= Decimal('0'):  # Long trade
                note = ''
                income = round(trade.close_amount(no_settlement=ns), 2)
                income_rub = round(trade.close_amount(self._currency_id, no_settlement=ns), 2)
                spending = round(trade.open_amount(no_settlement=ns), 2) + trade.fee()
                spending_rub = round(trade.open_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2)
            else:                            # Short trade
                # Check were there any dividends during short position holding
                short_dividend = Decimal('0')
                short_dividend_rub = Decimal('0')
                div_list = [x for x in dividends_withdrawn if x.asset().id() == trade.asset().id()]
                div_list = [x for x in div_list if trade.open_operation().settlement() <= x.ex_date() <= trade.close_operation().settlement()]
                for dividend in div_list:
                    short_dividend -= dividend.amount()
                    short_dividend_rub -= dividend.amount(self._currency_id)  # amount is negative
                dividends_withdrawn = [x for x in dividends_withdrawn if not x in div_list]
                note = f"Удержан дивиденд: {short_dividend_rub:.2f} RUB ({short_dividend:.2f} {self.account_currency.symbol()})" if short_dividend_rub > Decimal('0') else ''
                income = round(trade.open_amount(no_settlement=ns), 2)
                income_rub = round(trade.open_amount(self._currency_id, no_settlement=ns), 2)
                spending = round(trade.close_amount(no_settlement=ns), 2) + trade.fee() + short_dividend
                spending_rub = round(trade.close_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2) + short_dividend_rub
            line = {
                'report_template': "trade",
                'symbol': trade.asset().symbol(self.account_currency.id()),
                'isin': trade.asset().isin(),  # May be not used in template (for derivatives as example)
                'qty': trade.qty(),
                'country_iso': self.account.country().iso_code(),  # this field is required for DLSG
                'o_type': "Покупка" if trade.qty() >= Decimal('0') else "Продажа",
                'o_number': trade.open_operation().number(),
                'o_date': trade.open_operation().timestamp(),
                'o_rate': self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1],
                'os_date': trade.open_operation().settlement(),
                'os_rate': os_rate,
                'o_price': trade.open_operation().price(),
                'o_amount':  round(trade.open_amount(no_settlement=ns), 2),
                'o_amount_rub': round(trade.open_amount(self._currency_id, no_settlement=ns), 2),
                'o_fee': trade.open_fee(),
                'o_fee_rub': round(trade.open_fee(self._currency_id), 2),
                'c_type': "Продажа" if trade.qty() >= Decimal('0') else "Покупка",
                'c_number': trade.close_operation().number(),
                'c_date': trade.close_operation().timestamp(),
                'c_rate': self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1],
                'cs_date': trade.close_operation().settlement(),
                'cs_rate': cs_rate,
                'c_price': trade.close_operation().price(),
                'c_amount': round(trade.close_amount(no_settlement=ns), 2),
                'c_amount_rub': round(trade.close_amount(self._currency_id, no_settlement=ns), 2),
                'c_fee': trade.close_fee(),
                'c_fee_rub': round(trade.close_fee(self._currency_id), 2),
                'income': income,    # this field is required for DLSG
                'income_rub': income_rub,
                'spending_rub': spending_rub,
                'profit': income - spending,
                'profit_rub': income_rub - spending_rub,
                'note': note
            }
            deals_report.append(line)
        self.insert_totals(deals_report, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return deals_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_stocks_and_etf(self):
        trades = self.shares_trades_list()
        return self.prepare_trades_report(trades)

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_bonds(self):
        country = self.account.country()
        bonds_report = []
        ns = not self.use_settlement
        trades = self.bonds_trades_list()
        for trade in trades:
            if ns:
                os_rate = self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1]
            else:
                os_rate = self.account_currency.quote(trade.open_operation().settlement(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().settlement(), self._currency_id)[1]
            if trade.qty() >= Decimal('0'):  # Long trade
                income = round(trade.close_amount(no_settlement=ns), 2) + trade.close_operation().accrued_interest()
                income_rub = round(trade.close_amount(self._currency_id, no_settlement=ns), 2) + round(trade.close_operation().accrued_interest(self._currency_id), 2)
                spending = round(trade.open_amount(no_settlement=ns), 2) + trade.fee() - trade.open_operation().accrued_interest()
                spending_rub = round(trade.open_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2) - round(trade.open_operation().accrued_interest(self._currency_id), 2)
            else:                            # Short trade
                income = round(trade.open_amount(no_settlement=ns), 2) + trade.open_operation().accrued_interest()
                income_rub = round(trade.open_amount(self._currency_id, no_settlement=ns), 2) + round(trade.open_operation().accrued_interest(self._currency_id), 2)
                spending = round(trade.close_amount(no_settlement=ns), 2) + trade.fee() - trade.close_operation().accrued_interest()
                spending_rub = round(trade.close_amount(self._currency_id, no_settlement=ns), 2) + round(trade.fee(self._currency_id), 2) - round(trade.close_operation().accrued_interest(self._currency_id), 2)
            line = {
                'report_template': "bond_trade",
                'symbol': trade.asset().symbol(self.account_currency.id()),
                'isin': trade.asset().isin(),
                'qty': trade.qty(),
                'principal': trade.asset().principal(),
                'country_iso': country.iso_code(),
                'o_type': "Покупка" if trade.qty() >= Decimal('0') else "Продажа",
                'o_number': trade.open_operation().number(),
                'o_date': trade.open_operation().timestamp(),
                'o_rate': self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1],
                'os_date': trade.open_operation().settlement(),
                'os_rate': os_rate,
                'o_price': Decimal('100') * trade.open_operation().price() / trade.asset().principal(),
                'o_int': -trade.open_operation().accrued_interest(),
                'o_int_rub': -round(trade.open_operation().accrued_interest(self._currency_id), 2),
                'o_amount':  round(trade.open_amount(no_settlement=ns), 2),
                'o_amount_rub': round(trade.open_amount(self._currency_id, no_settlement=ns), 2),
                'o_fee': trade.open_fee(),
                'o_fee_rub': round(trade.open_fee(self._currency_id), 2),
                'c_type': "Продажа" if trade.qty() >= Decimal('0') else "Покупка",
                'c_number': trade.close_operation().number(),
                'c_date': trade.close_operation().timestamp(),
                'c_rate': self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1],
                'cs_date': trade.close_operation().settlement(),
                'cs_rate': cs_rate,
                'c_price': Decimal('100') * trade.close_operation().price() / trade.asset().principal(),
                'c_int': trade.close_operation().accrued_interest(),
                'c_int_rub': round(trade.close_operation().accrued_interest(self._currency_id), 2),
                'c_amount': round(trade.close_amount(no_settlement=ns), 2),
                'c_amount_rub': round(trade.close_amount(self._currency_id, no_settlement=ns), 2),
                'c_fee': trade.close_fee(),
                'c_fee_rub': round(trade.close_fee(self._currency_id), 2),
                'income': income,  # this field is required for DLSG
                'income_rub': income_rub,
                'spending_rub': spending_rub,
                'profit': income - spending,
                'profit_rub': income_rub - spending_rub
            }
            bonds_report.append(line)
        # Second - take all bond interest payments not linked with buy/sell transactions
        currency = JalAsset(self.account.currency())
        country = self.account.country()
        interests = Dividend.get_list(self.account.id(), subtype=Dividend.BondInterest, skip_accrued=True)
        interests = [x for x in interests if self.year_begin <= x.timestamp() <= self.year_end]  # Only in given range
        for interest in interests:
            amount = interest.amount()
            rate = currency.quote(interest.timestamp(), self._currency_id)[1]
            amount_rub = round(amount * rate, 2)
            line = {
                'report_template': "bond_interest",
                'type': "Купон",
                'empty': '',  # to keep cell borders drawn
                'o_date': interest.timestamp(),
                'symbol': interest.asset().symbol(currency.id()),
                'isin': interest.asset().isin(),
                'number': interest.number(),
                'interest': amount,
                'rate': rate,
                'interest_rub': amount_rub,
                'income_rub': amount_rub,
                'spending_rub': Decimal('0.0'),
                'profit': amount,
                'profit_rub': amount_rub,
                'country_iso': country.iso_code(),
            }
            bonds_report.append(line)
        self.insert_totals(bonds_report, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return bonds_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_derivatives(self):
        trades = self.derivatives_trades_list()
        return self.prepare_trades_report(trades)

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_crypto(self):
        country = self.account.country()
        crypto_report = []
        trades = self.account.closed_trades_list()
        trades = [x for x in trades if x.asset().type() == PredefinedAsset.Crypto]
        trades = [x for x in trades if x.close_operation().type() == LedgerTransaction.Trade]
        trades = [x for x in trades if x.open_operation().type() == LedgerTransaction.Trade]
        trades = [x for x in trades if self.year_begin <= x.close_operation().settlement() <= self.year_end]
        for trade in trades:
            o_rate = self.account_currency.quote(trade.open_operation().timestamp(), self._currency_id)[1]
            c_rate = self.account_currency.quote(trade.close_operation().timestamp(), self._currency_id)[1]
            if self.use_settlement:
                os_rate = self.account_currency.quote(trade.open_operation().settlement(), self._currency_id)[1]
                cs_rate = self.account_currency.quote(trade.close_operation().settlement(), self._currency_id)[1]
            else:
                os_rate = o_rate
                cs_rate = c_rate
            o_amount = round(trade.open_operation().price() * abs(trade.qty()), 2)
            o_amount_rub = round(o_amount * os_rate, 2)
            c_amount = round(trade.close_operation().price() * abs(trade.qty()), 2)
            c_amount_rub = round(c_amount * cs_rate, 2)
            o_fee = trade.open_operation().fee() * abs(trade.qty() / trade.open_operation().qty())
            c_fee = trade.close_operation().fee() * abs(trade.qty() / trade.close_operation().qty())
            income = c_amount if trade.qty() >= Decimal('0') else o_amount
            income_rub = c_amount_rub if trade.qty() >= Decimal('0') else o_amount_rub
            spending = o_amount if trade.qty() >= Decimal('0') else c_amount
            spending += o_fee + c_fee
            spending_rub = o_amount_rub if trade.qty() >= Decimal('0') else c_amount_rub
            spending_rub += round(o_fee * o_rate, 2) + round(c_fee * c_rate, 2)
            line = {
                'report_template': "trade",
                'symbol': trade.asset().symbol(self.account_currency.id()),
                'qty': trade.qty(),
                'country_iso': country.iso_code(),
                'o_type': "Покупка" if trade.qty() >= Decimal('0') else "Продажа",
                'o_number': trade.open_operation().number(),
                'o_date': trade.open_operation().timestamp(),
                'o_rate': o_rate,
                'os_date': trade.open_operation().settlement(),
                'os_rate': os_rate,
                'o_price': trade.open_operation().price(),
                'o_amount': o_amount,
                'o_amount_rub': o_amount_rub,
                'o_fee': o_fee,
                'o_fee_rub': round(o_fee * o_rate, 2),
                'c_type': "Продажа" if trade.qty() >= Decimal('0') else "Покупка",
                'c_number': trade.close_operation().number(),
                'c_date': trade.close_operation().timestamp(),
                'c_rate': c_rate,
                'cs_date': trade.close_operation().settlement(),
                'cs_rate': cs_rate,
                'c_price': trade.close_operation().price(),
                'c_amount': c_amount,
                'c_amount_rub': c_amount_rub,
                'c_fee': c_fee,
                'c_fee_rub': round(c_fee * c_rate, 2),
                'income': income,    # this field is required for DLSG
                'income_rub': income_rub,
                'spending_rub': spending_rub,
                'profit': income - spending,
                'profit_rub': income_rub - spending_rub
            }
            crypto_report.append(line)
        self.insert_totals(crypto_report, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return crypto_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_fees(self):
        fees_report = []
        fee_operations = JalCategory(PredefinedCategory.Fees).get_operations(self.year_begin, self.year_end)
        fee_operations = [x for x in fee_operations if x.account_id() == self.account.id()]
        for operation in fee_operations:
            rate = self.account_currency.quote(operation.timestamp(), self._currency_id)[1]
            fees = [x for x in operation.lines() if x['category_id'] == PredefinedCategory.Fees]
            for fee in fees:
                amount = -Decimal(fee['amount'])
                line = {
                    'report_template': "fee",
                    'payment_date': operation.timestamp(),
                    'rate': rate,
                    'amount': amount,
                    'amount_rub': round(amount * rate, 2),
                    'note': fee['note']
                }
                fees_report.append(line)
        self.insert_totals(fees_report, ["amount", "amount_rub"])
        return fees_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_interest(self):
        interests_report = []
        interest_operations = JalCategory(PredefinedCategory.Interest).get_operations(self.year_begin, self.year_end)
        interest_operations = [x for x in interest_operations if x.account_id() == self.account.id()]
        for operation in interest_operations:
            rate = self.account_currency.quote(operation.timestamp(), self._currency_id)[1]
            interests = [x for x in operation.lines() if x['category_id'] == PredefinedCategory.Interest]
            for interest in interests:
                amount = Decimal(interest['amount'])
                line = {
                    'report_template': "interest",
                    'payment_date': operation.timestamp(),
                    'rate': rate,
                    'amount': amount,
                    'amount_rub': round(amount * rate, 2),
                    'tax_rub': round(Decimal('0.13') * amount * rate, 2),
                    'note': interest['note']
                }
                interests_report.append(line)
        # Process cash payments out of corporate actions
        payments = CorporateAction.get_payments(self.account)
        payments = [x for x in payments if self.year_begin <= x['timestamp'] <= self.year_end]
        for payment in payments:
            rate = self.account_currency.quote(payment['timestamp'], self._currency_id)[1]
            line = {
                'report_template': "interest",
                'payment_date': payment['timestamp'],
                'rate': rate,
                'amount': payment['amount'],
                'amount_rub': round(payment['amount'] * rate, 2),
                'tax_rub': round(Decimal('0.13') * payment['amount'] * rate, 2),
                'note': payment['note']
            }
            interests_report.append(line)
        self.insert_totals(interests_report, ["amount", "amount_rub", "tax_rub"])
        return interests_report

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_corporate_actions(self):
        corporate_actions_report = []
        trades = self.account.closed_trades_list()
        trades = [x for x in trades if x.close_operation().type() == LedgerTransaction.Trade]
        trades = [x for x in trades if x.open_operation().type() == LedgerTransaction.CorporateAction]
        trades = [x for x in trades if self.year_begin <= x.close_operation().settlement() <= self.year_end]
        trades = sorted(trades, key=lambda x: (x.asset().symbol(self.account_currency.id()), x.close_operation().timestamp()))
        group = 1
        share = Decimal('1.0')   # This will track share of processed asset, so it starts from 100.0%
        previous_symbol = ""
        for trade in trades:
            lines = []
            sale = trade.close_operation()
            t_rate = self.account_currency.quote(sale.timestamp(), self._currency_id)[1]
            if self.use_settlement:
                s_rate = self.account_currency.quote(sale.settlement(), self._currency_id)[1]
            else:
                s_rate = t_rate
            if previous_symbol != sale.asset().symbol(self.account_currency.id()):
                # Clean processed qty records if symbol have changed
                self._processed_trade_qty = {}
                if sale.settlement() >= self.year_begin:  # Don't put sub-header of operation is out of scope
                    corporate_actions_report.append({
                        'report_template': "symbol_header",
                        'report_group': 0,
                        'description': f"Сделки по бумаге: {sale.asset().symbol(self.account_currency.id())} - {sale.asset().name()}"
                    })
                    previous_symbol = sale.asset().symbol(self.account_currency.id())
            amount = round(sale.price() * trade.qty(), 2)
            amount_rub = round(amount * s_rate, 2)
            fee_rub = round(sale.fee() * t_rate, 2)
            if sale.timestamp() < self.year_begin:    # Don't show deal that is before report year (level = -1)
                self.proceed_corporate_action(lines, trade, trade.qty(), share, -1, group)
            else:
                lines.append({
                    'report_template': "trade",
                    'report_group': group,
                    'operation': "Продажа",
                    't_date': sale.timestamp(),
                    't_rate': t_rate,
                    's_date': sale.settlement(),
                    's_rate': s_rate,
                    'symbol': sale.asset().symbol(self.account_currency.id()),
                    'isin': sale.asset().isin(),
                    'trade_number': sale.number(),
                    'price': sale.price(),
                    'qty': trade.qty(),
                    'amount': amount,
                    'amount_rub': amount_rub,
                    'fee': sale.fee(),
                    'fee_rub': fee_rub,
                    'basis_ratio': Decimal('100') * share,
                    'income_rub': amount_rub,
                    'spending_rub': fee_rub
                })
                if sale.asset().type() == PredefinedAsset.Bond:
                    self.output_accrued_interest(lines, sale, 1, 0)
                self.proceed_corporate_action(lines, trade, trade.qty(), share, 1, group)
            self.insert_totals(lines, ["income_rub", "spending_rub"])
            corporate_actions_report += lines
            group += 1
        return corporate_actions_report

    # actions - mutable list of tax records to output into json-report
    # trade - JalClosedTrade object, for which we need to proceed with opening corporate action
    # qty - amount of asset to process
    # share - value share that is being processed currently
    # level - how deep we are in a chain of events (is used for correct indents)
    # group - use for odd/even lines grouping in the report
    def proceed_corporate_action(self, actions, trade, qty, share, level, group):
        asset_id, qty, share = self.output_corp_action(actions, trade.open_operation(), trade.asset(), qty, share, level, group)
        next_level = -1 if level == -1 else (level + 1)
        self.next_corporate_action(actions, trade, qty, share, next_level, group)

    def next_corporate_action(self, actions, trade, qty, share, level, group):
        # get list of deals that were closed as result of current corporate action
        trades = self.account.closed_trades_list()
        trades = [x for x in trades if x.close_operation().type() == LedgerTransaction.CorporateAction]
        trades = [x for x in trades if x.close_operation().id() == trade.open_operation().id()]
        for item in trades:
            if item.open_operation().type() == LedgerTransaction.Trade:
                qty = self.output_purchase(actions, item.open_operation(), qty, share, level, group)
            elif item.open_operation().type() == LedgerTransaction.CorporateAction:
                self.proceed_corporate_action(actions, item, qty, share, level, group)
            else:
                assert False, "Unexpected opening transaction"

    def output_purchase(self, actions, purchase, proceed_qty, share, level, group):
        if proceed_qty <= Decimal('0'):
            return proceed_qty
        t_rate = self.account_currency.quote(purchase.timestamp(), self._currency_id)[1]
        if self.use_settlement:
            s_rate = self.account_currency.quote(purchase.settlement(), self._currency_id)[1]
        else:
            s_rate = t_rate
        if purchase.id() in self._processed_trade_qty:   # we have some qty processed already
            qty = purchase.qty() - self._processed_trade_qty[purchase.id()]
        else:
            qty = purchase.qty()
        if qty <= Decimal('0'):
            return proceed_qty   # This trade was fully matched during previous operations processing
        deal_qty = qty
        qty = proceed_qty if proceed_qty < deal_qty else deal_qty
        amount = round(purchase.price() * qty, 2)
        amount_rub = round(amount * s_rate, 2)
        fee = purchase.fee() * qty / deal_qty
        fee_rub = round(fee * t_rate, 2)
        # Update processed quantity for current _purchase_ operation
        self._processed_trade_qty[purchase.id()] = self._processed_trade_qty.get(purchase.id(), 0) + qty
        if level >= 0:  # Don't output if level==-1, i.e. corp action is out of report scope
            actions.append({
                'report_template': "trade",
                'report_group': group,
                'operation': ' ' * level * 3 + "Покупка",
                'trade_number': purchase.number(),
                'symbol': purchase.asset().symbol(self.account_currency.id()),
                'isin': purchase.asset().isin(),
                't_date': purchase.timestamp(),
                't_rate': t_rate,
                's_date': purchase.settlement(),
                's_rate': s_rate,
                'basis_ratio': Decimal('100') * share,
                'qty': qty,
                'price': purchase.price(),
                'amount': amount,
                'amount_rub': amount_rub,
                'fee': fee,
                'fee_rub': fee_rub,
                'income_rub': Decimal('0'),
                'spending_rub': round(share *(amount_rub + fee_rub), 2)
            })
        if purchase.asset().type() == PredefinedAsset.Bond:
            share = qty / deal_qty if qty < deal_qty else 1
            self.output_accrued_interest(actions, purchase, share, level)
        return proceed_qty - qty

    # asset - is a resulting asset that is being processed at current stage
    def output_corp_action(self, actions, action, asset, proceed_qty, share, level, group):
        currency = JalAsset(self.account.currency())
        if proceed_qty <= 0:
            return proceed_qty, share
        r_qty, r_share = action.get_result_for_asset(asset)
        share = share * r_share
        qty_before = action.qty() * proceed_qty / r_qty
        if action.subtype() == CorporateAction.SpinOff:
            action_results = action.get_results()
            spinoff = [x for x in action_results if x['asset_id'] != action.asset().id()]
            assert len(spinoff) == 1, "Multiple assets for spin-off"
            spinoff = spinoff[0]
            new_asset = JalAsset(spinoff['asset_id'])
            old_asset_name = f"{action.asset().symbol(currency.id())} ({action.asset().isin()})"
            new_asset_name = f"{new_asset.symbol(currency.id())} ({new_asset.isin()})"
            display_share = Decimal('100') * Decimal(spinoff['value_share'])
        else:
            old_asset_name = f"{action.asset().symbol(currency.id())} ({action.asset().isin()})"
            new_asset_name = f"{asset.symbol(currency.id())} ({asset.isin()})"
            display_share = Decimal('100') * r_share
        note = self.CorpActionText[action.subtype()].format(old=old_asset_name, new=new_asset_name,
                                                            before=remove_exponent(qty_before),
                                                            after=remove_exponent(proceed_qty), share=display_share)
        if level >= 0:  # Don't output if level==-1, i.e. corp action is out of report scope
            actions.append({
                'report_template': "action",
                'report_group': group,
                'operation': ' ' * level * 3 + "Корп. действие",
                'action_date': action.timestamp(),
                'action_number': action.number(),
                'symbol': action.asset().symbol(currency.id()),
                'isin': action.asset().isin(),
                'qty': action.qty(),
                'description': note
            })
        return action.asset().id(), qty_before, share

    def output_accrued_interest(self, actions, operation, share, level):
        country = self.account.country()
        accrued_interest = operation.accrued_interest()
        if accrued_interest == Decimal('0'):
            return
        rate = self.account_currency.quote(operation.timestamp(), self._currency_id)[1]
        interest = accrued_interest if share == 1 else share * accrued_interest
        interest_rub = abs(round(interest * rate, 2))
        if interest < 0:  # Accrued interest paid for purchase
            interest = -interest
            op_name = ' ' * level * 3 + "НКД уплачен"
            spending_rub = interest_rub
            income_rub = Decimal('0')
        else:                         # Accrued interest received for sale
            op_name = ' ' * level * 3 + "НКД получен"
            income_rub = interest_rub
            spending_rub = Decimal('0')
        actions.append({
            'report_template': 'bond_interest',
            'empty': '',
            'operation': op_name,
            'symbol': operation.asset().symbol(self.account_currency.id()),
            'isin': operation.asset().isin(),
            'number': operation.number(),
            'o_date': operation.timestamp(),
            'rate': rate,
            'interest': interest,
            'interest_rub': interest_rub,
            'income_rub': income_rub,
            'spending_rub': spending_rub,
            'country_iso': country.iso_code()
        })



from PySide6.QtCore import Slot
from PySide6.QtWidgets import QDialog, QMessageBox
from jal.ui.ui_select_reference_dlg import Ui_SelectReferenceDlg
from jal.widgets.helpers import center_window
from jal.widgets.reference_selector import PeerSelector
from jal.widgets.reference_selector import CategorySelector
from jal.widgets.reference_selector import TagSelector


#-----------------------------------------------------------------------------------------------------------------------
# Common base GUI dialog class for selector dialogs. Takes window title and label comment to describe selection
class SelectReferenceDialog(QDialog):
    def __init__(self, parent=None, title='', description=''):
        super().__init__(parent=parent)
        self.ui = Ui_SelectReferenceDlg()
        self.ui.setupUi(self)
        self.selected_id = 0
        self.setWindowTitle(title)
        self.ui.DescriptionLabel.setText(description)
        center_window(self)

    @Slot()
    def closeEvent(self, event):
        if self.selected_id == 0:
            QMessageBox().warning(None, self.tr("No selection"), self.tr("You should select something"), QMessageBox.Ok)
            event.ignore()
            return
        self.setResult(QDialog.Accepted)
        event.accept()


#-----------------------------------------------------------------------------------------------------------------------
# Dialog for peer selection
# Constructor takes description to show and default_peer for initial choice
class SelectPeerDialog(SelectReferenceDialog):
    def __init__(self, description, default_peer=0):
        super().__init__(title=self.tr("Please select peer"), description=description)
        self.PeerWidget = PeerSelector(self.ui.SelectorFrame)
        self.ui.FrameLayout.addWidget(self.PeerWidget)
        self.PeerWidget.selected_id = self.selected_id = default_peer

    @Slot()
    def closeEvent(self, event):
        self.selected_id = self.PeerWidget.selected_id
        super().closeEvent(event)
        
        
#-----------------------------------------------------------------------------------------------------------------------
# Dialog for category selection
# Constructor takes description to show and default_category for initial choice
class SelectCategoryDialog(SelectReferenceDialog):
    def __init__(self, parent=None, description='', default_category=0):
        super().__init__(parent, title=self.tr("Please select category"), description=description)
        self.CategoryWidget = CategorySelector(self.ui.SelectorFrame)
        self.ui.FrameLayout.addWidget(self.CategoryWidget)
        self.CategoryWidget.selected_id = self.selected_id = default_category

    @Slot()
    def closeEvent(self, event):
        self.selected_id = self.CategoryWidget.selected_id
        super().closeEvent(event)


#-----------------------------------------------------------------------------------------------------------------------
# Dialog for tag selection
# Constructor takes description to show and default_tag for initial choice
class SelectTagDialog(SelectReferenceDialog):
    def __init__(self, parent=None, description='', default_tag=0):
        super().__init__(parent, title=self.tr("Please select tag"), description=description)
        self.TagWidget = TagSelector(self.ui.SelectorFrame)
        self.ui.FrameLayout.addWidget(self.TagWidget)
        self.TagWidget.selected_id = self.selected_id = default_tag

    @Slot()
    def closeEvent(self, event):
        self.selected_id = self.TagWidget.selected_id
        super().closeEvent(event)



from __future__ import annotations
from typing import Optional
from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex

# ----------------------------------------------------------------------------------------------------------------------
# Base class to provide a common functionality of a tree element
class AbstractTreeItem:
    def __init__(self, parent=None, group=''):
        self._parent = parent
        self._children = []
        self._group = group

    def setParent(self, parent: AbstractTreeItem):
        self._parent = parent

    def getParent(self) -> AbstractTreeItem:
        return self._parent

    def childrenCount(self) -> int:
        return len(self._children)

    def appendChild(self, child: AbstractTreeItem):
        child.setParent(self)
        self._children.append(child)
        if self.isGroup():
            self.updateGroupDetails(child.details())

    def getChild(self, id) -> Optional[AbstractTreeItem]:
        if id < 0 or id >= len(self._children):
            return None
        return self._children[id]

    def details(self, **kwargs):
        raise NotImplementedError("To be defined in derived class")

    # Update current group data after new child with child_data was added into the group
    def updateGroupDetails(self, child_data):
        self._calculateGroupTotals(child_data)
        if self._parent is not None:
            self._parent.updateGroupDetails(child_data)
        self.updateChildrenData()

    # Update group children data (usually after some changes in group data)
    def updateChildrenData(self):
        if self.isGroup():
            for child in self._children:
                child._afterParentGroupUpdate(self.details())
                child.updateChildrenData()

    # This method is called if current item is a group. It should update group data with new child_data
    def _calculateGroupTotals(self, child_data):
        raise NotImplementedError("To be defined in derived class")

    # This method is called if parent group data were updated. It should update element data with new group_data
    def _afterParentGroupUpdate(self, group_data):
        raise NotImplementedError("To be defined in derived class")

    def isGroup(self) -> bool:
        if self._parent is None:   # Root element is always a group
            return True
        return self._group != ''


# ----------------------------------------------------------------------------------------------------------------------
# Base class to provide a common functionality of a tree model
class AbstractTreeModel(QAbstractItemModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._view = parent_view
        self._root = None
        self._groups = []
        self._columns = []

    def index(self, row, column, parent=None):
        if not parent.isValid():
            parent = self._root
        else:
            parent = parent.internalPointer()
        child = parent.getChild(row)
        if child:
            return self.createIndex(row, column, child)
        return QModelIndex()

    def parent(self, index=None):
        if not index.isValid():
            return QModelIndex()
        child_item = index.internalPointer()
        parent_item = child_item.getParent()
        if parent_item == self._root:
            return QModelIndex()
        return self.createIndex(0, 0, parent_item)

    def rowCount(self, parent=None):
        if not parent.isValid():
            parent_item = self._root
        else:
            parent_item = parent.internalPointer()
        if parent_item is not None:
            return parent_item.childrenCount()
        else:
            return 0

    def headerWidth(self, section):
        return self._view.header().sectionSize(section)

    def fieldIndex(self, field_name: str) -> int:
        for i, column in enumerate(self._columns):
            if column['field'] == field_name:
                return i
        return -1

    def columnCount(self, parent=None):
        if parent is None:
            parent_item = self._root
        elif not parent.isValid():
            parent_item = self._root
        else:
            parent_item = parent.internalPointer()
        if parent_item is not None:
            return len(self._columns)
        else:
            return 0

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                return self._columns[section]['name']
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignCenter)
        return None

    # defines report grouping by provided field list - 'group_field1;group_field2;...'
    # return True if grouping was actually changed and False otherwise
    def setGrouping(self, group_list) -> bool:
        new_groups = group_list.split(';') if group_list else []
        if new_groups == self._groups:
            return False
        self._groups = new_groups
        return True

    def prepareData(self):
        raise NotImplementedError("To be defined in descendant class")

    def configureView(self):
        raise NotImplementedError("To be defined in descendant class")



from math import log10, floor, ceil
from decimal import Decimal

from PySide6.QtCore import Qt, QMargins, QDateTime, QDate
from PySide6.QtWidgets import QWidget, QHBoxLayout
from PySide6.QtCharts import QChartView, QLineSeries, QScatterSeries, QDateTimeAxis, QValueAxis
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.constants import CustomColor
from jal.widgets.mdi import MdiWidget


class ChartWidget(QWidget):
    def __init__(self, parent, quotes, trades, data_range, currency_name):
        super().__init__(parent=parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)

        self.quotes_series = QLineSeries()
        for point in quotes:            # Conversion to 'float' in order not to get 'int' overflow on some platforms
            self.quotes_series.append(float(point['timestamp']), point['quote'])

        self.trade_series = QScatterSeries()
        for point in trades:            # Conversion to 'float' in order not to get 'int' overflow on some platforms
            self.trade_series.append(float(point['timestamp']), point['price'])
        self.trade_series.setMarkerSize(5)
        self.trade_series.setBorderColor(CustomColor.LightRed)
        self.trade_series.setBrush(CustomColor.DarkRed)

        axisX = QDateTimeAxis()
        axisX.setTickCount(11)
        axisX.setRange(QDateTime().fromSecsSinceEpoch(data_range[0]), QDateTime().fromSecsSinceEpoch(data_range[1]))
        axisX.setFormat("yyyy/MM/dd")
        axisX.setLabelsAngle(-90)
        axisX.setTitleText("Date")

        axisY = QValueAxis()
        axisY.setTickCount(11)
        axisY.setRange(data_range[2], data_range[3])
        axisY.setTitleText("Price, " + currency_name)

        self.chartView = QChartView()
        self.chartView.chart().addSeries(self.quotes_series)
        self.chartView.chart().addSeries(self.trade_series)
        self.chartView.chart().addAxis(axisX, Qt.AlignBottom)
        self.chartView.chart().setAxisX(axisX, self.quotes_series)
        self.chartView.chart().setAxisX(axisX, self.trade_series)
        self.chartView.chart().addAxis(axisY, Qt.AlignLeft)
        self.chartView.chart().setAxisY(axisY, self.quotes_series)
        self.chartView.chart().setAxisY(axisY, self.trade_series)
        self.chartView.chart().legend().hide()
        self.chartView.setViewportMargins(0, 0, 0, 0)
        self.chartView.chart().layout().setContentsMargins(0, 0, 0, 0)  # To remove extra spacing around chart
        self.chartView.chart().setBackgroundRoundness(0)  # To remove corner rounding
        self.chartView.chart().setMargins(QMargins(0, 0, 0, 0))  # Allow chart to fill all space

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove extra space around layout
        self.layout.addWidget(self.chartView)
        self.setLayout(self.layout)


class ChartWindow(MdiWidget):
    def __init__(self, account_id, asset_id, currency_id, _asset_qty, parent=None):
        super().__init__(parent)

        self.account_id = account_id
        self.asset_id = asset_id
        self.currency_id = currency_id if asset_id != currency_id else 1  # Check whether we have currency or asset
        self.asset_name = JalAsset(self.asset_id).symbol(JalAccount(self.account_id).currency())
        self.quotes = []
        self.trades = []
        self.currency_name = ''
        self.range = [0, 0, 0, 0]

        self.prepare_chart_data()

        self.chart = ChartWidget(self, self.quotes, self.trades, self.range, self.currency_name)

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Remove extra space around layout
        self.layout.addWidget(self.chart)
        self.setLayout(self.layout)

        self.setWindowTitle(self.tr("Price chart for ") + self.asset_name)

        self.ready = True

    def prepare_chart_data(self):
        account = JalAccount(self.account_id)
        asset = JalAsset(self.asset_id)
        self.currency_name = JalAsset(account.currency()).symbol()
        positions = account.open_trades_list(asset)
        start_time = min([x['operation'].timestamp() for x in positions]) - 2592000  # Shift back by 30 days
        quotes = asset.quotes(start_time, QDate.currentDate().endOfDay(Qt.UTC).toSecsSinceEpoch(), self.currency_id)
        for quote in quotes:
            self.quotes.append({'timestamp': quote[0] * 1000, 'quote': quote[1]})  # timestamp to ms
        for trade in positions:
            self.trades.append({
                'timestamp': trade['operation'].timestamp() * 1000,  # timestamp to ms
                'price': trade['price'],
                'qty': trade['remaining_qty']
            })
        if self.quotes or self.trades:
            min_price = min([x['quote'] for x in self.quotes] + [x['price'] for x in self.trades])
            max_price = max([x['quote'] for x in self.quotes] + [x['price'] for x in self.trades])
            min_ts = min([x['timestamp'] for x in self.quotes] + [x['timestamp'] for x in self.trades]) / 1000
            max_ts = max([x['timestamp'] for x in self.quotes] + [x['timestamp'] for x in self.trades]) / 1000
        else:
            self.range = [0, 0, 0, 0]
            return
        # push range apart if we have very close points
        if min_price == max_price:
            min_price = Decimal('0.95') * min_price
            max_price = Decimal('1.05') * max_price
        if min_ts == max_ts:
            min_ts = 0.95 * min_ts
            max_ts = 1.05 * max_ts
        # Round min/max values to near "round" values in order to have 10 nice intervals
        step = Decimal(10) ** Decimal(floor(Decimal.log10(max_price - min_price)))
        min_price = floor(min_price / step) * step
        max_price = ceil(max_price / step) * step
        # Add a gap at the beginning and end
        min_ts = int(min_ts - 86400 * 3)
        max_ts = int(max_ts + 86400 * 3)
        self.range = [min_ts, max_ts, min_price, max_price]



from PySide6.QtCore import Signal, Slot, Property
from PySide6.QtWidgets import QDialog, QWidget, QPushButton, QComboBox, QMenu, QHBoxLayout, QCheckBox, QMessageBox
from jal.db.db import JalModel
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.widgets.helpers import center_window
from jal.widgets.reference_dialogs import AccountListDialog
from jal.ui.ui_select_account_dlg import Ui_SelectAccountDlg


########################################################################################################################
#  UI Button to choose accounts
########################################################################################################################
class AccountButton(QPushButton):
    changed = Signal(int)

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.p_account_id = 0

        self.Menu = QMenu(self)
        self.Menu.addAction(self.tr("Choose account"), self.ChooseAccount)
        self.Menu.addAction(self.tr("Any account"), self.ClearAccount)
        self.setMenu(self.Menu)

        self.dialog = AccountListDialog()
        self.setText(self.dialog.SelectedName)

    def getId(self):
        return self.p_account_id

    def setId(self, account_id):
        self.p_account_id = account_id
        if self.p_account_id:
            self.setText(JalAccount(account_id).name())
        else:
            self.setText(self.tr("ANY"))
        self.changed.emit(self.p_account_id)

    account_id = Property(int, getId, setId, notify=changed)

    def ChooseAccount(self):
        ref_point = self.mapToGlobal(self.geometry().bottomLeft())
        self.dialog.setGeometry(ref_point.x(), ref_point.y(), self.dialog.width(), self.dialog.height())
        self.dialog.setFilter()
        res = self.dialog.exec(enable_selection=True)
        if res:
            self.account_id = self.dialog.selected_id

    def ClearAccount(self):
        self.account_id = 0


#-----------------------------------------------------------------------------------------------------------------------
# Dialog for account selection
# Constructor takes description to show and recent_account for default choice.
# Selected account won't be equal to current_account
class SelectAccountDialog(QDialog):
    def __init__(self, description, current_account, recent_account=None):
        super().__init__()
        self.ui = Ui_SelectAccountDlg()
        self.ui.setupUi(self)
        self.account_id = recent_account
        self.store_account = False
        self.current_account = current_account

        self.ui.DescriptionLbl.setText(description)
        if self.account_id:
            self.ui.AccountWidget.selected_id = self.account_id
        center_window(self)

    @Slot()
    def closeEvent(self, event):
        self.account_id = self.ui.AccountWidget.selected_id
        self.store_account = self.ui.ReuseAccount.isChecked()
        if self.ui.AccountWidget.selected_id == 0:
            QMessageBox().warning(None, self.tr("No selection"), self.tr("Invalid account selected"), QMessageBox.Ok)
            event.ignore()
            return

        if self.ui.AccountWidget.selected_id == self.current_account:
            QMessageBox().warning(None, self.tr("No selection"), self.tr("Please select different account"),
                                  QMessageBox.Ok)
            event.ignore()
            return

        self.setResult(QDialog.Accepted)
        event.accept()


# ----------------------------------------------------------------------------------------------------------------------
class CurrencyComboBox(QComboBox):
    changed = Signal(int)

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.p_selected_id = 0
        self.model = None
        self.activated.connect(self.OnUserSelection)

        self.model = JalModel(self, "currencies")
        self.model.select()
        self.setModel(self.model)
        self.setModelColumn(self.model.fieldIndex("symbol"))

    def getId(self):
        return self.p_selected_id

    def setId(self, new_id):
        if self.p_selected_id == new_id:
            return
        self.p_selected_id = new_id
        name = JalAsset(self.p_selected_id).symbol()
        if self.currentIndex() == self.findText(name):
            return
        self.setCurrentIndex(self.findText(name))

    selected_id = Property(int, getId, setId, notify=changed, user=True)

    def setIndex(self, index):
        if index is not None:
            self.selected_id = index
            self.changed.emit(self.selected_id)

    @Slot()
    def OnUserSelection(self, _selected_index):
        self.selected_id = self.model.record(self.currentIndex()).value("id")
        self.changed.emit(self.selected_id)


# ----------------------------------------------------------------------------------------------------------------------
class OptionalCurrencyComboBox(QWidget):
    changed = Signal()
    name_updated = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._id = 0

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.null_flag = QCheckBox(parent)
        self.null_flag.setChecked(False)
        self.null_flag.setText(self.tr("Currency"))
        self.layout.addWidget(self.null_flag)
        self.currency = CurrencyComboBox(parent)
        self.currency.setEnabled(False)
        self.layout.addWidget(self.currency)
        self.setLayout(self.layout)

        self.setFocusProxy(self.null_flag)
        self.null_flag.clicked.connect(self.onClick)
        self.currency.changed.connect(self.onCurrencyChange)

    def setText(self, text):
        self.null_flag.setText(text)

    def getId(self):
        return self._id if self._id else None

    def setId(self, new_value):
        if self._id == new_value:
            return
        self._id = new_value
        self.updateView()
        name = JalAsset(self._id).symbol()
        self.name_updated.emit('' if name is None else name)

    currency_id = Property(int, getId, setId, notify=changed, user=True)

    def updateView(self):
        has_value = True if self._id else False
        if has_value:
            self.currency.selected_id = self._id
        self.null_flag.setChecked(has_value)
        self.currency.setEnabled(has_value)

    @Slot()
    def onClick(self):
        if self.null_flag.isChecked():
            if self.currency.selected_id == 0:
                self.currency.selected_id = JalAsset.get_base_currency()
            self.currency_id = self.currency.selected_id
        else:
            self.currency_id = 0
        self.changed.emit()

    @Slot()
    def onCurrencyChange(self, _id):
        self.currency_id = self.currency.selected_id
        self.changed.emit()



from decimal import Decimal
from PySide6.QtCore import Qt, Slot, QDate, QAbstractTableModel, QModelIndex
from PySide6.QtGui import QBrush, QFontDatabase
from PySide6.QtWidgets import QStyledItemDelegate, QHeaderView
from jal.constants import CustomColor
from jal.db.ledger import Ledger
from jal.db.operations import LedgerTransaction
from jal.widgets.helpers import ts2dt


class OperationsModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [" ", self.tr("Timestamp"), self.tr("Account"), self.tr("Notes"),
                         self.tr("Amount"), self.tr("Balance"), self.tr("Currency")]
        self._view = parent_view
        self._amount_delegate = None
        self._data = []
        self._begin = 0
        self._end = 0
        self._account = 0
        self._bold_font = QFontDatabase.systemFont(QFontDatabase.GeneralFont)
        self._bold_font.setBold(True)

        self.prepareData()

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def get_operation(self, row):
        if (row >= 0) and (row < self.rowCount()):
            return self._data[row]['op_type'], self._data[row]['id']
        else:
            return [0, 0]

    def data(self, index, role=Qt.DisplayRole, field=''):
        row = index.row()
        if not index.isValid():
            return None
        operation = LedgerTransaction().get_operation(self._data[row]['op_type'],
                                                      self._data[row]['id'],
                                                      self._data[row]['subtype'])
        if role == Qt.DisplayRole:
            return self.data_text(operation, index.column())
        if role == Qt.FontRole and index.column() == 0:
            # below line isn't related with font, it is put here to be called for each row minimal times (ideally 1)
            self._view.setRowHeight(row, self._view.verticalHeader().defaultSectionSize() * operation.view_rows())
            return self._bold_font
        if role == Qt.ForegroundRole and self._view.isEnabled():
            if index.column() == 0:
                return QBrush(operation.label_color())
            elif index.column() == 5:
                if operation.reconciled():
                    return QBrush(CustomColor.Blue)
        if role == Qt.TextAlignmentRole:
            if index.column() == 0:
                return int(Qt.AlignCenter | Qt.AlignVCenter)
            if index.column() == 4 or index.column() == 5:
                return int(Qt.AlignRight)
            return int(Qt.AlignLeft)
        if role == Qt.UserRole:  # return underlying data for given field extra parameter
            return self._data[index.row()][field]

    def data_text(self, operation, column):
        if column == 0:
            return operation.label()
        elif column == 1:
            date_time = ts2dt(operation.timestamp())
            if operation.number():
                date_time += f"\n# {operation.number()}"
            return date_time
        elif column == 2:
            if operation.asset_name():
                return operation.account_name() + "\n" + operation.asset_name()
            else:
                return operation.account_name()
        elif column == 3:
            return operation.description()
        elif column == 4:
            return operation.value_change()
        elif column == 5:
            return operation.value_total()
        elif column == 6:
            return operation.value_currency()
        else:
            assert False, "Unexpected column number"

    def configureView(self):
        self._view.setColumnWidth(0, 10)
        self._view.setColumnWidth(1, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        self._view.setColumnWidth(2, 300)
        self._view.horizontalHeader().setSectionResizeMode(3, QHeaderView.Stretch)
        self._view.horizontalHeader().setFont(self._bold_font)
        self._amount_delegate = ColoredAmountsDelegate(self._view)
        self._view.setItemDelegateForColumn(4, self._amount_delegate)

        self._view.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)  # row size is adjusted in data() method

    @Slot()
    def setAccount(self, account_id):
        if self._account != account_id:
            self._account = account_id
            self.prepareData()

    def getAccount(self):
        return self._account

    @Slot()
    def setDateRange(self, start, end=0):
        self._begin = start
        if end:
            self._end = end
        else:
            self._end = QDate.currentDate().endOfDay(Qt.UTC).toSecsSinceEpoch()
        self.prepareData()

    def update(self):
        self.prepareData()

    @Slot()
    def refresh(self):
        idx = self._view.selectionModel().selection().indexes()
        idx = idx[0] if idx else QModelIndex()  # Take first selected or empty index
        if self._view.model() != self:          # View uses some proxy model
            idx = self._view.model().mapToSource(idx)
        self.prepareData()
        if self._view.model() != self:
            idx = self._view.model().mapFromSource(idx)
        self._view.setCurrentIndex(idx)

    def prepareData(self):
        self._data = []
        self._data = Ledger.get_operations_sequence(self._begin, self._end, self._account)
        self.modelReset.emit()

    def deleteRows(self, rows):
        for row in rows:
            if (row >= 0) and (row < len(self._data)):
                LedgerTransaction.get_operation(self._data[row]['op_type'], self._data[row]['id'],
                                                display_type=self._data[row]['subtype']).delete()
        self.prepareData()


# ----------------------------------------------------------------------------------------------------------------------
class ColoredAmountsDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        self._view = parent
        super().__init__(parent=parent)

    def paint(self, painter, option, index):
        painter.save()
        data = index.model().data(index)
        rect = option.rect
        H = rect.height()
        Y = rect.top()
        rect.setHeight(H / len(data))
        for i, item in enumerate(data):
            rect.moveTop(Y + i * (H / len(data)))
            self.draw_value(option.rect, painter, item)
        painter.restore()

    def draw_value(self, rect, painter, value):
        if value is None:
            return
        pen = painter.pen()
        try:
            if self._view.isEnabled():
                if value >= 0:
                    pen.setColor(CustomColor.DarkGreen)
                else:
                    pen.setColor(CustomColor.DarkRed)
            painter.setPen(pen)
            painter.drawText(rect, Qt.AlignRight | Qt.AlignVCenter, f"{value:+,.2f}")
            if abs(value - round(value, 2)) > Decimal('0'):  # Underline decimal part
                shift = painter.fontMetrics().horizontalAdvance(f"{value:+,.2f}"[-2:])
                painter.drawLine(rect.right() - shift, rect.bottom(), rect.right(), rect.bottom())
        except TypeError:
            pass



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QMainWindow, QMenu, QMenuBar,
    QSizePolicy, QSplitter, QStatusBar, QVBoxLayout,
    QWidget)

from jal.widgets.custom.log_viewer import LogViewer
from jal.widgets.mdi import TabbedMdiArea

class Ui_JAL_MainWindow(object):
    def setupUi(self, JAL_MainWindow):
        if not JAL_MainWindow.objectName():
            JAL_MainWindow.setObjectName(u"JAL_MainWindow")
        JAL_MainWindow.resize(835, 436)
        JAL_MainWindow.setMinimumSize(QSize(0, 0))
        self.actionExit = QAction(JAL_MainWindow)
        self.actionExit.setObjectName(u"actionExit")
        self.actionExit.setMenuRole(QAction.QuitRole)
        self.action_Re_build_Ledger = QAction(JAL_MainWindow)
        self.action_Re_build_Ledger.setObjectName(u"action_Re_build_Ledger")
        self.action_LoadQuotes = QAction(JAL_MainWindow)
        self.action_LoadQuotes.setObjectName(u"action_LoadQuotes")
        self.actionAccounts = QAction(JAL_MainWindow)
        self.actionAccounts.setObjectName(u"actionAccounts")
        self.actionAssets = QAction(JAL_MainWindow)
        self.actionAssets.setObjectName(u"actionAssets")
        self.actionPeers = QAction(JAL_MainWindow)
        self.actionPeers.setObjectName(u"actionPeers")
        self.actionCategories = QAction(JAL_MainWindow)
        self.actionCategories.setObjectName(u"actionCategories")
        self.actionBackup = QAction(JAL_MainWindow)
        self.actionBackup.setObjectName(u"actionBackup")
        self.actionRestore = QAction(JAL_MainWindow)
        self.actionRestore.setObjectName(u"actionRestore")
        self.PrepareTaxForms = QAction(JAL_MainWindow)
        self.PrepareTaxForms.setObjectName(u"PrepareTaxForms")
        self.actionTags = QAction(JAL_MainWindow)
        self.actionTags.setObjectName(u"actionTags")
        self.actionImportSlipRU = QAction(JAL_MainWindow)
        self.actionImportSlipRU.setObjectName(u"actionImportSlipRU")
        self.actionQuotes = QAction(JAL_MainWindow)
        self.actionQuotes.setObjectName(u"actionQuotes")
        self.actionOperations = QAction(JAL_MainWindow)
        self.actionOperations.setObjectName(u"actionOperations")
        self.PrepareFlowReport = QAction(JAL_MainWindow)
        self.PrepareFlowReport.setObjectName(u"PrepareFlowReport")
        self.actionBaseCurrency = QAction(JAL_MainWindow)
        self.actionBaseCurrency.setObjectName(u"actionBaseCurrency")
        self.actionCleanAll = QAction(JAL_MainWindow)
        self.actionCleanAll.setObjectName(u"actionCleanAll")
        self.centralwidget = QWidget(JAL_MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.centralwidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.splitter = QSplitter(self.centralwidget)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Vertical)
        self.mdiArea = TabbedMdiArea(self.splitter)
        self.mdiArea.setObjectName(u"mdiArea")
        self.splitter.addWidget(self.mdiArea)
        self.Logs = LogViewer(self.splitter)
        self.Logs.setObjectName(u"Logs")
        self.splitter.addWidget(self.Logs)

        self.verticalLayout.addWidget(self.splitter)

        JAL_MainWindow.setCentralWidget(self.centralwidget)
        self.MainMenu = QMenuBar(JAL_MainWindow)
        self.MainMenu.setObjectName(u"MainMenu")
        self.MainMenu.setGeometry(QRect(0, 0, 835, 23))
        self.menuMain = QMenu(self.MainMenu)
        self.menuMain.setObjectName(u"menuMain")
        self.menu_Data = QMenu(self.MainMenu)
        self.menu_Data.setObjectName(u"menu_Data")
        self.menu_Export = QMenu(self.MainMenu)
        self.menu_Export.setObjectName(u"menu_Export")
        self.menuLanguage = QMenu(self.MainMenu)
        self.menuLanguage.setObjectName(u"menuLanguage")
        self.menuImport = QMenu(self.MainMenu)
        self.menuImport.setObjectName(u"menuImport")
        self.menuStatement = QMenu(self.menuImport)
        self.menuStatement.setObjectName(u"menuStatement")
        self.menuReports = QMenu(self.MainMenu)
        self.menuReports.setObjectName(u"menuReports")
        JAL_MainWindow.setMenuBar(self.MainMenu)
        self.StatusBar = QStatusBar(JAL_MainWindow)
        self.StatusBar.setObjectName(u"StatusBar")
        JAL_MainWindow.setStatusBar(self.StatusBar)

        self.MainMenu.addAction(self.menuMain.menuAction())
        self.MainMenu.addAction(self.menu_Data.menuAction())
        self.MainMenu.addAction(self.menuReports.menuAction())
        self.MainMenu.addAction(self.menuImport.menuAction())
        self.MainMenu.addAction(self.menu_Export.menuAction())
        self.MainMenu.addAction(self.menuLanguage.menuAction())
        self.menuMain.addAction(self.actionOperations)
        self.menuMain.addAction(self.actionExit)
        self.menu_Data.addSeparator()
        self.menu_Data.addAction(self.action_Re_build_Ledger)
        self.menu_Data.addSeparator()
        self.menu_Data.addAction(self.actionAccounts)
        self.menu_Data.addAction(self.actionAssets)
        self.menu_Data.addAction(self.actionPeers)
        self.menu_Data.addAction(self.actionCategories)
        self.menu_Data.addAction(self.actionTags)
        self.menu_Data.addAction(self.actionQuotes)
        self.menu_Data.addSeparator()
        self.menu_Data.addAction(self.actionBaseCurrency)
        self.menu_Data.addSeparator()
        self.menu_Data.addAction(self.actionBackup)
        self.menu_Data.addAction(self.actionRestore)
        self.menu_Data.addSeparator()
        self.menu_Data.addAction(self.actionCleanAll)
        self.menu_Export.addAction(self.PrepareTaxForms)
        self.menu_Export.addAction(self.PrepareFlowReport)
        self.menuImport.addAction(self.action_LoadQuotes)
        self.menuImport.addAction(self.menuStatement.menuAction())
        self.menuImport.addAction(self.actionImportSlipRU)

        self.retranslateUi(JAL_MainWindow)

        QMetaObject.connectSlotsByName(JAL_MainWindow)
    # setupUi

    def retranslateUi(self, JAL_MainWindow):
        JAL_MainWindow.setWindowTitle(QCoreApplication.translate("JAL_MainWindow", u"jal", None))
        self.actionExit.setText(QCoreApplication.translate("JAL_MainWindow", u"&Exit", None))
        self.action_Re_build_Ledger.setText(QCoreApplication.translate("JAL_MainWindow", u"Re-build &Ledger...", None))
        self.action_LoadQuotes.setText(QCoreApplication.translate("JAL_MainWindow", u"&Quotes...", None))
        self.actionAccounts.setText(QCoreApplication.translate("JAL_MainWindow", u"&Accounts", None))
        self.actionAssets.setText(QCoreApplication.translate("JAL_MainWindow", u"A&ssets", None))
        self.actionPeers.setText(QCoreApplication.translate("JAL_MainWindow", u"&Peers", None))
        self.actionCategories.setText(QCoreApplication.translate("JAL_MainWindow", u"&Categories", None))
        self.actionBackup.setText(QCoreApplication.translate("JAL_MainWindow", u"&Backup...", None))
        self.actionRestore.setText(QCoreApplication.translate("JAL_MainWindow", u"&Restore...", None))
        self.PrepareTaxForms.setText(QCoreApplication.translate("JAL_MainWindow", u"Investment &tax report", None))
        self.actionTags.setText(QCoreApplication.translate("JAL_MainWindow", u"&Tags", None))
        self.actionImportSlipRU.setText(QCoreApplication.translate("JAL_MainWindow", u"Slip [RU]...", None))
        self.actionQuotes.setText(QCoreApplication.translate("JAL_MainWindow", u"&Quotes", None))
        self.actionOperations.setText(QCoreApplication.translate("JAL_MainWindow", u"&Operations", None))
        self.PrepareFlowReport.setText(QCoreApplication.translate("JAL_MainWindow", u"Foreign accounts &flow report", None))
        self.actionBaseCurrency.setText(QCoreApplication.translate("JAL_MainWindow", u"&Base Currency", None))
        self.actionCleanAll.setText(QCoreApplication.translate("JAL_MainWindow", u"Clean All", None))
        self.menuMain.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Main", None))
        self.menu_Data.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Data", None))
        self.menu_Export.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Export", None))
        self.menuLanguage.setTitle(QCoreApplication.translate("JAL_MainWindow", u"L&anguage", None))
        self.menuImport.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Import", None))
        self.menuStatement.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Statement", None))
        self.menuReports.setTitle(QCoreApplication.translate("JAL_MainWindow", u"&Reports", None))
    # retranslateUi




import os
import logging
from jal.constants import CustomColor
from jal.db.helpers import load_icon
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QApplication, QPlainTextEdit, QLabel, QPushButton
from PySide6.QtGui import QBrush, QAction


# Adapter class to have custom log handler that may be passed to logger.addHandler/logger.removeHandler methods and
# then forward all messages parent view to display them
class LogHandler(logging.Handler):
    def __init__(self, parent_view):
        self._parent_view = parent_view
        super().__init__()

    def emit(self, record, **kwargs):
        message = self.format(record)
        self._parent_view.displayMessage(record.levelno, message)


# A GUI class to display messages from python logging unit in a normal multi-line text area
class LogViewer(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()
        self._logger = None     # Here an instance of current logger will be stored
        self._log_handler = LogHandler(self)
        self.setReadOnly(True)
        self.status_bar = None    # Status bar where notifications and control are located
        self.expandButton = None  # Button that shows/hides log window
        self.notification = None  # Here is QLabel element to display LOG update status
        self.clear_color = None   # Variable to store initial "clear" background color
        self.collapsed_text = self.tr("▶ logs")
        self.expanded_text = self.tr("▲ logs")

        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.addAction(load_icon("copy.png"), self.tr('Copy'), self._copy2clipboard)
        self.addAction(self.tr('Select all'), self.selectAll)
        self.addAction(load_icon("delete.png"), self.tr('Clear'), self.clear)

    def _copy2clipboard(self):
        cursor = self.textCursor()
        text = cursor.selectedText() if cursor.selectedText().toPlainText() else self.toPlainText()
        QApplication.clipboard().setText(text)

    def startLogging(self):
        self._logger = logging.getLogger()
        self._logger.addHandler(self._log_handler)
        log_level = os.environ.get('LOGLEVEL', 'INFO').upper()
        self._logger.setLevel(log_level)
        self._log_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))

    def stopLogging(self):
        self._logger.removeHandler(self._log_handler)    # Removing handler (but it doesn't prevent exception at exit)
        logging.raiseExceptions = False                  # Silencing logging module exceptions

    def displayMessage(self, level: int, message: str):
        predefinded_colors = {
            logging.DEBUG: CustomColor.Grey,
            logging.INFO: self.clear_color,
            logging.WARNING: CustomColor.LightRed,
            logging.ERROR: CustomColor.LightRed,
            logging.CRITICAL: CustomColor.LightRed
        }
        try:
            msg_color = predefinded_colors[level]
        except KeyError:
            self.appendPlainText(self.tr("Unknown logging level provided: ") + f"{level}")
            msg_color = CustomColor.LightRed

        # Store message in log window
        tf = self.currentCharFormat()
        tf.setForeground(QBrush(msg_color))
        self.setCurrentCharFormat(tf)
        self.appendPlainText(message)

        # Show in status bar
        if self.notification:
            palette = self.notification.palette()
            palette.setColor(self.notification.foregroundRole(), msg_color)
            self.notification.setPalette(palette)
            msg = message.replace('\n', "; ")  # Get rid of new lines in error message
            elided_text = self.notification.fontMetrics().elidedText(msg, Qt.ElideRight, self.get_available_width())
            self.notification.setText(elided_text)
        # Set button color
        if self.expandButton:
            palette = self.expandButton.palette()
            palette.setColor(self.expandButton.foregroundRole(), msg_color)
        self.app.processEvents()

    def showEvent(self, event):
        self.cleanNotification()
        super().showEvent(event)

    def setStatusBar(self, status_bar):
        self.setVisible(False)
        self.status_bar = status_bar

        self.expandButton = QPushButton(self.collapsed_text, parent=self)
        self.expandButton.setFixedWidth(self.expandButton.fontMetrics().horizontalAdvance(self.collapsed_text) * 1.25)
        self.expandButton.setCheckable(True)
        self.expandButton.clicked.connect(self.showLogs)
        self.status_bar.addWidget(self.expandButton)

        self.notification = QLabel(self)
        self.status_bar.addWidget(self.notification)
        self.notification.setAutoFillBackground(True)
        self.clear_color = self.expandButton.palette().color(self.notification.foregroundRole())

    def removeStatusBar(self):
        self.cleanNotification()
        self.notification = None

    def cleanNotification(self):
        if self.notification:
            palette = self.notification.palette()
            palette.setColor(self.notification.foregroundRole(), self.clear_color)
            self.notification.setPalette(palette)
            self.notification.setText("")

    @Slot()
    def showLogs(self):
        self.setVisible(self.expandButton.isChecked())
        text = self.expanded_text if self.expandButton.isChecked() else self.collapsed_text
        self.expandButton.setText(text)

    # Calculates maximum width that is free on status bar
    def get_available_width(self):
        width = self.status_bar.width()
        for child in self.status_bar.children():
            if hasattr(child, "width") and child != self.notification:
                width -= child.width()
        return width - 8    # return calculated width reduced by small safety gap



import logging
import re
from datetime import datetime, timezone

from jal.constants import Setup, PredefinedCategory
from jal.data_import.statement import FOF, Statement_ImportError
from jal.data_import.statement_xls import StatementXLS

JAL_STATEMENT_CLASS = "StatementUKFU"


# ----------------------------------------------------------------------------------------------------------------------
class StatementUKFU(StatementXLS):
    Header = (0, 0, '  Брокер: ООО "Твой Брокер"')
    PeriodPattern = (0, 2, r"  за период с (?P<S>\d\d\.\d\d\.\d\d\d\d) по (?P<E>\d\d\.\d\d\.\d\d\d\d)")
    AccountPattern = (2, 6, None)
    SummaryHeader = "СОСТОЯНИЕ ДЕНЕЖНЫХ СРЕДСТВ НА СЧЕТЕ"
    trade_columns = {
        "number": "Номер сделки",
        "date": "Дата сделки",
        "time": "Время сделки",
        "settlement": "Дата поставки, плановая",
        "isin": "ISIN",
        "B/S": "Вид сделки",
        "price": "Цена одной ЦБ",
        "qty": "Количество ЦБ, шт.",
        "amount": "Сумма сделки",
        "accrued_int": "НКД",
        "fee_ex": "Комиссия ТС",
        "currency": "Валюта цены"
    }
    money_section = "ПОЗИЦИЯ ПО ДЕНЕЖНЫМ СРЕДСТВАМ"
    money_columns = {
        "name": "Код Валюты",
        "begin": "Входящий остаток",
        "end": "Плановый исходящий остаток",
        "settled_end": "Исходящий остаток"
    }
    asset_section = "СОСТОЯНИЕ ПОРТФЕЛЯ ЦЕННЫХ БУМАГ"
    asset_columns = {
        "name": "Наименование ЦБ",
        "isin": "ISIN",
        "reg_number": "Номер гос. регистрации / CFI код"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("Tvoy Broker (ex. Uralsib Broker)")
        self.icon_name = "uralsib.ico"
        self.filename_filter = self.tr("Tvoy Broker statement (*.zip)")
        self.asset_withdrawal = []

    def _load_deals(self):
        self.load_stock_deals()
        self.load_futures_deals()
        self.load_asset_movements()

    def _load_cash_transactions(self):
        self.load_cash_transactions()
        self.load_broker_fee()

    def load_stock_deals(self):
        cnt = 0
        columns = {
            "number": "Номер сделки",
            "date": "Дата сделки",
            "time": "Время сделки",
            "isin": "ISIN",
            "B/S": "Вид сделки",
            "price": "Цена одной ЦБ",
            "currency": "Валюта цены",
            "qty": r"Количество ЦБ, шт\.",
            "amount": "Сумма сделки",
            "accrued_int": "НКД",
            "settlement": "Дата поставки, плановая",
            "fee_ex": "Комиссия ТС"
        }

        row, headers = self.find_section_start("СДЕЛКИ С ЦЕННЫМИ БУМАГАМИ", columns,
                                               subtitle="Биржевые сделки с ценными бумагами в отчетном периоде",
                                               header_height=2)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '' and self._statement[self.HeaderCol][row + 1] == '':
                break
            if self._statement[self.HeaderCol][row].startswith('Итого по выпуску:') or \
                    self._statement[self.HeaderCol][row] == '':
                row += 1
                continue
            try:
                deal_number = int(self._statement[self.HeaderCol][row])
            except ValueError:
                row += 1
                continue
            try:
                code = self.currency_id(self.currency_substitutions[self._statement[headers['currency']][row]])
            except KeyError:
                code = self.currency_id(self._statement[headers['currency']][row])
            asset_id = self.asset_id({'isin': self._statement[headers['isin']][row],
                                      'currency': code, 'search_online': "MOEX"})
            if self._statement[headers['B/S']][row] == 'Покупка':
                qty = self._statement[headers['qty']][row]
                bond_interest = -self._statement[headers['accrued_int']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
                bond_interest = self._statement[headers['accrued_int']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue

            price = self._statement[headers['price']][row]
            currency = self._statement[headers['currency']][row]
            fee = self._statement[headers['fee_ex']][row]
            amount = self._statement[headers['amount']][row]
            if abs(abs(price * qty) - amount) >= self.RU_PRICE_TOLERANCE:
                price = abs(amount / qty)
            ts_string = self._statement[headers['date']][row] + ' ' + self._statement[headers['time']][row]
            timestamp = int(datetime.strptime(ts_string, "%d.%m.%Y %H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
            settlement = int(datetime.strptime(self._statement[headers['settlement']][row],
                                               "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, currency)
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": str(deal_number), "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            if bond_interest != 0:
                new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
                payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                           "number": str(deal_number), "asset": asset_id, "amount": bond_interest, "description": "НКД"}
                self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Trades loaded: ") + f"{cnt}")

    def load_futures_deals(self):
        cnt = 0
        columns = {
            "number": "Номер сделки",
            "date": "Дата сделки",
            "time": "Время сделки",
            "symbol": "Код контракта",
            "B/S": "Вид сделки",
            "price": "Цена фьючерса",
            "currency": "Валюта цены",
            "qty": r"Количество контрактов, шт\.",
            "amount": "Сумма",
            "settlement": "Дата расчетов по сделке",
            "fee_broker": r"Комиссия брокера, руб\.",
            "fee_ex": r"Комиссия ТС, руб\."
        }

        row, headers = self.find_section_start("СДЕЛКИ С ФЬЮЧЕРСАМИ И ОПЦИОНАМИ", columns,
                                               subtitle="Сделки с фьючерсами")
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '' and self._statement[self.HeaderCol][row + 1] == '':
                break
            if self._statement[self.HeaderCol][row].startswith("Входящая позиция по контракту") or \
                    self._statement[self.HeaderCol][row].startswith("Итого по контракту") or \
                    self._statement[self.HeaderCol][row] == '':
                row += 1
                continue
            try:
                deal_number = int(self._statement[self.HeaderCol][row])
            except ValueError:
                row += 1
                continue
            try:
                code = self.currency_id(self.currency_substitutions[self._statement[headers['currency']][row]])
            except KeyError:
                code = self.currency_id(self._statement[headers['currency']][row])
            asset_id = self.asset_id({'symbol': self._statement[headers['symbol']][row],
                                      'currency': code, 'search_online': "MOEX"})
            if self._statement[headers['B/S']][row] == 'Покупка':
                qty = self._statement[headers['qty']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue

            price = self._statement[headers['price']][row]
            currency = self._statement[headers['currency']][row]
            fee = self._statement[headers['fee_broker']][row] + self._statement[headers['fee_ex']][row]
            amount = self._statement[headers['amount']][row]
            if abs(abs(price * qty) - amount) >= self.RU_PRICE_TOLERANCE:
                price = abs(amount / qty)
            ts_string = self._statement[headers['date']][row] + ' ' + self._statement[headers['time']][row]
            timestamp = int(datetime.strptime(ts_string, "%d.%m.%Y %H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
            settlement = int(datetime.strptime(self._statement[headers['settlement']][row],
                                               "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, currency)
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            cnt += 1
            row += 1
        logging.info(self.tr("Futures trades loaded: ") + f"{cnt}")

    def load_asset_movements(self):
        cnt = 0
        columns = {
            "number": "№ операции",
            "date": "Дата",
            "type": "Тип операции",
            "asset": "Наименование ЦБ",
            "reg_number": r"Номер гос\. регистрации",
            "qty": "Количество ЦБ",
            "note": "Комментарий"
        }
        operations = {
            'Списание ЦБ после погашения': self.asset_cancellation,
            'Перевод ЦБ': self.asset_transfer,
            'Ввод ЦБ': self.asset_transfer_in,
            'Списано по сделке': None,  # These operations are results of trades
            'Получено по сделке': None,
        }

        row, headers = self.find_section_start("ДВИЖЕНИЕ ЦЕННЫХ БУМАГ ЗА ОТЧЕТНЫЙ ПЕРИОД", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol + 1][row] == '':  # Stop if there are no next date available
                break
            operation = self._statement[headers['type']][row]
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported asset operation ") + f"'{operation}'")
            number = self._statement[headers['number']][row]
            timestamp = int(datetime.strptime(self._statement[headers['date']][row],
                                              "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            asset = self.asset_id({'reg_number': self._statement[headers['reg_number']][row], 'search_online': "MOEX"})
            qty = self._statement[headers['qty']][row]
            description = self._statement[headers['note']][row]
            if operations[operation] is not None:
                operations[operation](timestamp, number, asset, qty, description)
            cnt += 1
            row += 1
        logging.info(self.tr("Asset operations loaded: ") + f"{cnt}")

    def asset_cancellation(self, timestamp, number, asset, qty, description):
        # Statement has negative value for cancellation - will be used to create sell trade
        record = {"timestamp": timestamp, "asset": asset, "number": number, "quantity": qty, "note": description}
        self.asset_withdrawal.append(record)

    def asset_transfer(self, timestamp, number, asset, qty, description):
        TransferPattern = r"^Перевод ЦБ на с\/с (?P<account_to>[\w|\/]+) с с\/с (?P<account_from>[\w|\/]+)\..*$"
        parts = re.match(TransferPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse asset transfer description ") + f"'{description}'")
        transfer = parts.groupdict()
        if len(transfer) != TransferPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Asset transfer description miss some data ") + f"'{description}'")
        currency_id = [x for x in self._data[FOF.SYMBOLS] if x["asset"] == asset][0]['currency']
        currency_name = [x for x in self._data[FOF.SYMBOLS] if x["asset"] == currency_id][0]['symbol']
        account_from = self._find_account_id(transfer['account_from'], currency_name)
        account_to = self._find_account_id(transfer['account_to'], currency_name)
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_from, account_to, 0], "asset": [asset, asset],
                    "timestamp": timestamp, "withdrawal": qty, "deposit": qty, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def asset_transfer_in(self, timestamp, number, asset, qty, description):
        currency_id = [x for x in self._data[FOF.SYMBOLS] if x["asset"] == asset][0]['currency']
        currency_name = [x for x in self._data[FOF.SYMBOLS] if x["asset"] == currency_id][0]['symbol']
        account_id = self._find_account_id(self._account_number, currency_name)
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0], "asset": [asset, asset],
                    "timestamp": timestamp, "withdrawal": qty, "deposit": qty, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def load_cash_transactions(self):
        cnt = 0
        columns = {
            "number": "№ операции",
            "date": "Дата",
            "type": "Тип операции",
            "amount": "Сумма",
            "currency": "Валюта",
            "description": "Комментарий"
        }
        operations = {
            'Ввод ДС': self.transfer_in,
            'Вывод ДС': self.transfer_out,
            'Перевод ДС': self.transfer,
            'Налог': self.tax,
            'Доход по финансовым инструментам': self.dividend,
            'Погашение купона': self.interest,
            'Погашение номинала': self.bond_repayment,
            'Списано по сделке': None,   # These operations are results of trades
            'Получено по сделке': None,
            "Вариационная маржа": None,  # These are non-trade operations for derivatives
            "Заблокировано средств ГО": None,
            "Разблокировано средств ГО": None,
            "Депозитарные сборы других депозитариев": self.fee
        }

        row, headers = self.find_section_start("ДВИЖЕНИЕ ДЕНЕЖНЫХ СРЕДСТВ ЗА ОТЧЕТНЫЙ ПЕРИОД",  columns)
        if row < 0:
            return False

        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol + 1][row] == '':  # Stop if there are no next date available
                break
            operation = self._statement[headers['type']][row]
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported cash transaction ") + f"'{operation}'")
            number = self._statement[headers['number']][row]
            timestamp = int(datetime.strptime(self._statement[headers['date']][row],
                                              "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            amount = self._statement[headers['amount']][row]
            description = self._statement[headers['description']][row]
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            if operations[operation] is not None:
                operations[operation](timestamp, number, account_id, amount, description)
            cnt += 1
            row += 1
        logging.info(self.tr("Cash operations loaded: ") + f"{cnt}")

    def transfer(self, timestamp, number, account_id, amount, description):
        TransferPattern = r"^Перевод ДС на с\/с (?P<account_to>[\w|\/]+) с с\/с (?P<account_from>[\w|\/]+)\..*$"
        if amount < 0:  # there should be positive paired record
            return
        parts = re.match(TransferPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse money transfer description ") + f"'{description}'")
        transfer = parts.groupdict()
        if len(transfer) != TransferPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Money transfer description miss some data ") + f"'{description}'")
        if transfer['account_from'] == transfer['account_to']:  # It is a technical record for incoming transfer
            return
        currency_id = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]['currency']
        currency_name = [x for x in self._data[FOF.SYMBOLS] if x["asset"] == currency_id][0]['symbol']
        account_from = self._find_account_id(transfer['account_from'], currency_name)
        account_to = self._find_account_id(transfer['account_to'], currency_name)
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_from, account_to, 0], "number": number,
                    "asset": [currency_id, currency_id], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_in(self, timestamp, number, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0], "number": number,
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, number, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0], "number": number,
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def dividend(self, timestamp, number, account_id, amount, description):
        DividendPattern = r"> (?P<DESCR1>.*) \((?P<REG_NUMBER>.*)\)((?P<DESCR2> .*)?(?P<TAX_TEXT> налог (в размере (?P<TAX>\d+\.\d\d) )?.*удержан))?\. НДС не облагается\."
        ISINPattern = r"[A-Z]{2}.{9}\d"

        parts = re.match(DividendPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse dividend description ") + f"'{description}'")
        dividend_data = parts.groupdict()
        isin_match = re.match(ISINPattern, dividend_data['REG_NUMBER'])
        currency_code = self.currency_id('RUB')
        if isin_match:
            asset_id = self.asset_id({'isin': dividend_data['REG_NUMBER'],
                                      'currency': currency_code, 'search_online': "MOEX"})
        else:
            asset_id = self.asset_id({'reg_number': dividend_data['REG_NUMBER'],
                                      'currency': currency_code, 'search_online': "MOEX"})
        if dividend_data['DESCR2']:
            short_description = dividend_data['DESCR1'] + ' ' + dividend_data['DESCR2'].strip()
        else:
            short_description = dividend_data['DESCR1']
        if dividend_data['TAX']:
            try:
                tax = float(dividend_data['TAX'])
            except ValueError:
                raise Statement_ImportError(self.tr("Failed to convert dividend tax ") + f"'{description}'")
        else:
            tax = 0
            if dividend_data['TAX_TEXT']:
                short_description += '; ' + dividend_data['TAX_TEXT'].strip()
        amount = amount + tax   # Statement contains value after taxation while JAL stores value before tax
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_DIVIDEND, "account": account_id, "timestamp": timestamp,
                   "number": number, "asset": asset_id, "amount": amount, "tax": tax, "description": short_description}
        self._data[FOF.ASSET_PAYMENTS].append(payment)

    def interest(self, timestamp, number, account_id, amount, description):
        BondInterestPattern = r"Погашение купона №( -?\d+)? (?P<NAME>.*)"

        parts = re.match(BondInterestPattern, description, re.IGNORECASE)
        if parts is None:
            logging.error(self.tr("Can't parse bond interest description ") + f"'{description}'")
            return
        interest_data = parts.groupdict()
        asset_id = self.asset_id({'symbol': interest_data['NAME'], 'should_exist': True})
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                   "number": number, "asset": asset_id, "amount": amount, "description": description}
        self._data[FOF.ASSET_PAYMENTS].append(payment)

    def bond_repayment(self, timestamp, _number, account_id, amount, description):
        BondRepaymentPattern = r"Погашение номинала (?P<NAME>.*)"

        parts = re.match(BondRepaymentPattern, description, re.IGNORECASE)
        if parts is None:
            logging.error(self.tr("Can't parse bond repayment description ") + f"'{description}'")
            return
        interest_data = parts.groupdict()
        asset_id = self.asset_id({'symbol': interest_data['NAME'], 'should_exist': True})
        match = [x for x in self.asset_withdrawal if x['asset'] == asset_id and x['timestamp'] == timestamp]
        if not match:
            logging.error(self.tr("Can't find asset cancellation record for ") + f"'{description}'")
            return
        if len(match) != 1:
            logging.error(self.tr("Multiple asset cancellation match for ") + f"'{description}'")
            return
        asset_cancel = match[0]

        qty = asset_cancel['quantity']
        price = abs(amount / qty)   # Price is always positive
        note = description + ", " + asset_cancel['note']
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
        trade = {"id": new_id, "number": asset_cancel['number'], "timestamp": timestamp, "settlement": timestamp,
                 "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": 0.0, "note": note}
        self._data[FOF.TRADES].append(trade)

    def tax(self, timestamp, _number, account_id, amount, description):
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        tax = {"id": new_id, "timestamp": timestamp, "account": account_id, "peer": 0,
               "lines": [{"amount": amount, "category": -PredefinedCategory.Taxes, "description": description}]}
        self._data[FOF.INCOME_SPENDING].append(tax)

    def fee(self, timestamp, _number, account_id, amount, description):
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        fee = {"id": new_id, "timestamp": timestamp, "account": account_id, "peer": 0,
               "lines": [{"amount": amount, "category": -PredefinedCategory.Fees, "description": description}]}
        self._data[FOF.INCOME_SPENDING].append(fee)

    def load_broker_fee(self):
        header_row = self.find_row(self.SummaryHeader) + 1
        if header_row < 0:
            logging.warning(self.tr("Can't get header to find fees"))
            return
        header_found = False
        for i, row in self._statement.iterrows():
            if (not header_found) and (row[self.HeaderCol] == "Уплаченная комиссия, в том числе"):
                header_found = True  # Start of broker fees list
                continue
            if header_found:
                if row[self.HeaderCol] != "":     # End of broker fee list
                    break
                for col in range(6, self._statement.shape[1]):
                    if not self._statement[col][header_row]:
                        break
                    try:
                        fee = float(row[col])
                    except (ValueError, TypeError):
                        continue
                    if fee == 0:
                        continue
                    if row[1] == 'комиссия торговой системы':  # Exchange fee is part of trades
                        continue
                    account_id = self._find_account_id(self._account_number, self._statement[col][header_row])
                    new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
                    fee = {"id": new_id, "timestamp": self._data[FOF.PERIOD][1], "account": account_id, "peer": 0,
                           "lines": [{"amount": fee, "category": -PredefinedCategory.Fees, "description": row[1]}]}
                    self._data[FOF.INCOME_SPENDING].append(fee)



import logging
from datetime import time, datetime, timedelta, timezone
from PySide6.QtGui import QImage
from PySide6.QtWidgets import QApplication
from jal.constants import Setup
try:
    from pyzbar import pyzbar
except ImportError:
    pass  # Helpers that use this imports shouldn't be called if imports are absent

# -----------------------------------------------------------------------------------------------------------------------
# Returns True if all modules from module_list are present in the system
def dependency_present(module_list: list) -> bool:
    result = True
    for module in module_list:
        try:
            __import__(module)
        except ImportError:
            result = False
    return result

# -----------------------------------------------------------------------------------------------------------------------
# Check if given signal of an object is connected or not
def is_signal_connected(object, signal_name) -> bool:
    meta_object = object.metaObject()
    method_index = meta_object.indexOfSignal(signal_name)
    return object.isSignalConnected(meta_object.method(method_index))

# -----------------------------------------------------------------------------------------------------------------------
# center given window with respect to main application window
def center_window(window):
    main_window = None
    for widget in QApplication.topLevelWidgets():
        if widget.objectName() == Setup.MAIN_WND_NAME:
            main_window = widget
    if main_window:
        x = main_window.x() + main_window.width() / 2 - window.width() / 2
        y = main_window.y() + main_window.height() / 2 - window.height() / 2
        window.setGeometry(x, y, window.width(), window.height())

# -----------------------------------------------------------------------------------------------------------------------
# converts string with leading zeros to integer (like '03'->3, '00'->0, ''->0)
def str2int(value: str) -> int:
    value = value.lstrip('0')
    if value:
        return int(value)
    else:
        return 0

# -----------------------------------------------------------------------------------------------------------------------
# converts given unix-timestamp into string that represents date and time
def ts2dt(timestamp: int) -> str:
    return datetime.utcfromtimestamp(timestamp).strftime('%d/%m/%Y %H:%M:%S')

# -----------------------------------------------------------------------------------------------------------------------
# converts given unix-timestamp into string that represents date
def ts2d(timestamp: int) -> str:
    return datetime.utcfromtimestamp(timestamp).strftime('%d/%m/%Y')

# -----------------------------------------------------------------------------------------------------------------------
# converts given datetime value into unix-timestamp
def dt2ts(value: datetime) -> int:
    return int(value.replace(tzinfo=timezone.utc).timestamp())

# -----------------------------------------------------------------------------------------------------------------------
# returns unix timestamp for the first second of the month/year
def month_start_ts(year: int, month: int) -> int:
    start = datetime(year=year, month=month, day=1, hour=0, minute=0, second=0)
    return dt2ts(start)

# returns unix timestamp for the first second of the given week of the month
def week_start_ts(year: int, week: int) -> int:
    start = datetime.strptime(f"{year}-W{week}-1", "%Y-W%W-%w")
    return dt2ts(start)

# -----------------------------------------------------------------------------------------------------------------------
# returns unix timestamp for the last second of the month/year
def month_end_ts(year: int, month: int) -> int:
    start = datetime(year=year, month=month, day=1, hour=0, minute=0, second=0)
    if month == 12:
        end = start.replace(year=year + 1, month=1)
    else:
        end = start.replace(month=month + 1)
    end = end - timedelta(seconds=1)
    return dt2ts(end)

# returns unix timestamp for the last second of the given week of the month
def week_end_ts(year: int, week: int) -> int:
    end = datetime.strptime(f"{year}-W{week}-1", "%Y-W%W-%w") + timedelta(days=7)
    return dt2ts(end)

# -----------------------------------------------------------------------------------------------------------------------
# returns a list of dictionaries for each month between 'begin' and 'end' timestamps (including):
# { year, number, begin_ts, end_ts }
# where number is a month number from 1 to 12
def month_list(begin: int, end: int) -> list:
    result = []
    year_begin = int(datetime.utcfromtimestamp(begin).strftime('%Y'))
    month_begin = str2int(datetime.utcfromtimestamp(begin).strftime('%m'))
    year_end = int(datetime.utcfromtimestamp(end).strftime('%Y'))
    month_end = str2int(datetime.utcfromtimestamp(end).strftime('%m'))
    for year in range(year_begin, year_end+1):
        month1 = month_begin if year == year_begin else 1
        month2 = month_end + 1 if year == year_end else 13
        for month in range(month1, month2):
            result.append({'year': year, 'number': month,
                           'begin_ts': month_start_ts(year, month), 'end_ts': month_end_ts(year, month)})
    return result

# returns a list of dictionaries for each week between 'begin' and 'end' timestamps (including):
# { year, number, begin_ts, end_ts }
# where number is a week number in year from 1 to 53
def week_list(begin: int, end: int) -> list:
    result = []
    year_begin = int(datetime.utcfromtimestamp(begin).strftime('%Y'))
    week_begin = str2int(datetime.utcfromtimestamp(begin).strftime('%W'))
    if week_begin == 0:
        year_begin -= 1
        week_begin = 53
    year_end = int(datetime.utcfromtimestamp(end).strftime('%Y'))
    week_end = int(datetime.utcfromtimestamp(end).strftime('%W'))
    if week_end == 0:
        year_end -= 1
        week_end = 53
    for year in range(year_begin, year_end + 1):
        wk1 = week_begin if year == year_begin else 1
        wk2 = week_end + 1 if year == year_end else 54
        for week in range(wk1, wk2):
            result.append({'year': year, 'number': week,
                           'begin_ts': week_start_ts(year, week), 'end_ts': week_end_ts(year, week)})
    return result

# ----------------------------------------------------------------------------------------------------------------------
# Function takes an image and searches for QR in it. Content of first found QR is returned. Otherwise - empty string.
def decodeQR(qr_image: QImage) -> str:
    if qr_image.isNull():
        return ''
    if not dependency_present(['pyzbar']):
        logging.warning("Package pyzbar not found for QR recognition.")
        return ''
    qr_image.convertTo(QImage.Format_Grayscale8)
    # bytesPerXXX is more accurate than width and height
    data = (qr_image.bits().tobytes(), qr_image.bytesPerLine(), int(qr_image.sizeInBytes()/qr_image.bytesPerLine()))
    barcodes = pyzbar.decode(data, symbols=[pyzbar.ZBarSymbol.QRCODE])
    if barcodes:
        return barcodes[0].data.decode('utf-8')
    return ''


# -----------------------------------------------------------------------------------------------------------------------
# Helpers to work with datetime
class ManipulateDate:
    @staticmethod
    def toTimestamp(date_value):
        time_value = time(0, 0, 0)
        dt_value = datetime.combine(date_value, time_value)
        return int(dt_value.replace(tzinfo=timezone.utc).timestamp())

    @staticmethod
    def PreviousWeek(day=datetime.today()):
        end = day + timedelta(days=1)
        prev_week = day - timedelta(days=7)
        start_of_week = prev_week - timedelta(days=prev_week.weekday())
        return ManipulateDate.toTimestamp(start_of_week), ManipulateDate.toTimestamp(end)

    @staticmethod
    def PreviousMonth(day=datetime.today()):
        end = day + timedelta(days=1)
        first_day_of_month = day.replace(day=1)
        last_day_of_prev_month = first_day_of_month - timedelta(days=1)
        first_day_of_prev_month = last_day_of_prev_month.replace(day=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_month), ManipulateDate.toTimestamp(end)

    @staticmethod
    def PreviousQuarter(day=datetime.today()):
        end = day + timedelta(days=1)
        prev_quarter_month = day.month - day.month % 3 - 3
        if prev_quarter_month > 0:
            quarter_back = day.replace(month=prev_quarter_month)
        else:
            quarter_back = day.replace(month=(prev_quarter_month + 12), year=(day.year - 1))
        first_day_of_prev_quarter = quarter_back.replace(day=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_quarter), ManipulateDate.toTimestamp(end)

    @staticmethod
    def PreviousYear(day=datetime.today()):
        end = day + timedelta(days=1)
        first_day_of_year = day.replace(day=1, month=1)
        last_day_of_prev_year = first_day_of_year - timedelta(days=1)
        first_day_of_prev_year = last_day_of_prev_year.replace(day=1, month=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_year), ManipulateDate.toTimestamp(end)

    @staticmethod
    def QuarterToDate(day=datetime.today()):
        end = day + timedelta(days=1)
        begin_month = day.month - 3
        if begin_month > 0:
            begin = day.replace(month=begin_month)
        else:
            begin = day.replace(month=(begin_month + 12), year=(day.year - 1))
        begin = begin.replace(day=1)
        return ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end)

    @staticmethod
    def YearToDate(day=datetime.today()):
        end = day + timedelta(days=1)
        begin = day.replace(day=1, year=(day.year - 1))
        return ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end)

    @staticmethod
    def ThisYear(day=datetime.today()):
        end = day + timedelta(days=1)
        begin = day.replace(day=1, month=1)
        return ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end)

    @staticmethod
    def LastYear(day=datetime.today()):
        end = day.replace(day=1, month=1)
        begin = end.replace(year=(day.year - 1))
        return ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end)

    @staticmethod
    def AllDates(day=datetime.today()):
        end = day.replace(day=1, month=1)
        return 0, ManipulateDate.toTimestamp(end)
